from typing import Any, Optional

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QRadioButton,
    QVBoxLayout,
    QWidget,
)

from first_breaks.desktop.utils import (
    TMappingSetup,
    validate_mapping_setup_and_get_current_index,
)


class QRadioSetWidget(QWidget):
    index_changed_signal = pyqtSignal(int)
    value_changed_signal = pyqtSignal(object)
    label_changed_signal = pyqtSignal(str)
    changed_signal = pyqtSignal()

    def __init__(
        self,
        mapping: TMappingSetup,
        current_index: Optional[int] = None,
        current_label: Optional[str] = None,
        current_value: Optional[str] = None,
        orientation: str = "horizontal",
        margins: Optional[int] = None,
        *args: Any,
        **kwargs: Any
    ):
        super().__init__(*args, **kwargs)
        assert orientation in ["horizontal", "vertical"]

        current_index = validate_mapping_setup_and_get_current_index(
            mapping,
            current_index=current_index,
            current_label=current_label,
            current_value=current_value,
        )

        self.mapping = mapping
        self.current_index = current_index

        if orientation == "horizontal":
            self.layout = QHBoxLayout()
        else:
            self.layout = QVBoxLayout()
        self.setLayout(self.layout)
        if margins is not None:
            self.layout.setContentsMargins(margins, margins, margins, margins)

        self.storage = {}
        self.selected_idx = current_index

        for idx, (label, value) in mapping.items():
            radio = QRadioButton(label)
            self.storage[radio] = idx
            if idx == current_index:
                radio.setChecked(True)
            radio.toggled.connect(lambda *lambda_args, **lambda_kwargs: self.changed())
            self.layout.addWidget(radio)

    def changed(self) -> None:
        radio = self.sender()
        if radio.isChecked():
            self.selected_idx = self.storage[radio]

            self.index_changed_signal.emit(self.selected_idx)
            self.value_changed_signal.emit(self.value())
            self.label_changed_signal.emit(self.label())
            self.changed_signal.emit()

    def value(self) -> Any:
        return self.mapping[self.selected_idx][1]

    def text(self) -> str:
        return str(self.value())

    def label(self) -> Any:
        return self.mapping[self.selected_idx][0]


if __name__ == "__main__":
    app = QApplication([])
    window = QRadioSetWidget({0: ("A", "aa"), 1: ("B", "bb"), 2: ("C", "cc")})
    window.show()
    app.exec_()



from typing import Any, List, Tuple

import pyqtgraph as pg
from PyQt5.QtCore import QEvent, QObject, QPointF, Qt, pyqtSignal

from first_breaks.desktop.utils import get_mouse_position_in_scene_coords
from first_breaks.utils.utils import generate_color


def get_rect_of_roi(roi: pg.ROI) -> Tuple[float, float, float, float]:
    x0, y0 = roi.pos().x(), roi.pos().y()
    x1, y1 = roi.pos().x() + roi.size().x(), roi.pos().y() + roi.size().y()
    x_min, x_max = sorted([x0, x1])
    y_min, y_max = sorted([y0, y1])
    return x_min, y_min, x_max, y_max


def invert_roi_shape(roi: pg.ROI) -> pg.ROI:
    position, size = roi.pos(), roi.size()
    roi.setPos((position.y(), position.x()))
    roi.setSize((size.y(), size.x()))
    return roi


class RoiManager(QObject):
    roi_added_signal = pyqtSignal(object)
    roi_change_started_signal = pyqtSignal(object)
    roi_change_finished_signal = pyqtSignal(object)
    roi_changing_signal = pyqtSignal(object)
    roi_hover_signal = pyqtSignal(object)
    roi_clicked_signal = pyqtSignal(object)
    roi_deleted_signal = pyqtSignal(object)

    def __init__(self, viewbox: pg.ViewBox):
        super().__init__()
        self.viewbox = viewbox
        self.rois: List[pg.ROI] = []
        self.selected_roi = None
        self.creating_roi = False

        self.viewbox.installEventFilter(self)
        self.viewbox.setFocusPolicy(Qt.ClickFocus)  # need to receive keyboard events

        self.mouse_move_signal = pg.SignalProxy(self.viewbox.scene().sigMouseMoved, rateLimit=60, slot=self.mouse_moved)

    def change_rois_view(self) -> None:
        for roi in self.rois:
            invert_roi_shape(roi)

    def mouse_moved(self, pos: Tuple[QPointF]) -> None:
        pos = pos[0]
        if self.creating_roi:
            self.update_roi_size(get_mouse_position_in_scene_coords(pos, self.viewbox))

    def eventFilter(self, obj: Any, event: QEvent) -> Any:
        # event_dict = {value: name for name, value in vars(QEvent).items() if isinstance(value, int)}
        # print(event.type(), event_dict.get(event.type(), "UNK"))
        if obj == self.viewbox:
            if event.type() == QEvent.GraphicsSceneMousePress:
                if event.button() == Qt.LeftButton and event.modifiers() == Qt.ShiftModifier:
                    self.creating_roi = True
                    self.start_roi_creation(get_mouse_position_in_scene_coords(event, self.viewbox))
                    return True
                self.select_roi()  # we want to accept other action by regular clicking, so we don't return True
            elif event.type() == QEvent.GraphicsSceneMouseRelease:
                if event.button() == Qt.LeftButton and self.creating_roi:
                    self.creating_roi = False
                    return True

            elif event.type() == QEvent.KeyPress:
                if event.key() == Qt.Key_Delete:
                    self.delete_selected_roi()
                    return True

        # If the event wasn't handled, propagate it further
        return super(RoiManager, self).eventFilter(obj, event)

    def start_roi_creation(self, position: QPointF) -> None:
        color = generate_color()
        pen = pg.mkPen(color=color, width=3)
        hover_pen = pg.mkPen(color=color, width=5)
        roi = pg.ROI([position.x(), position.y()], size=[0, 0], movable=True, pen=pen, hoverPen=hover_pen)

        roi.sigClicked.connect(self.on_roi_clicked)
        roi.sigRegionChangeStarted.connect(self.roi_change_started_signal)
        roi.sigRegionChanged.connect(self.roi_changing_signal)
        roi.sigRegionChangeFinished.connect(self.roi_change_finished_signal)
        roi.sigClicked.connect(self.roi_clicked_signal)
        self.viewbox.addItem(roi)
        self.rois.append(roi)
        self.roi_added_signal.emit(roi)
        self.selected_roi = roi
        self.creating_roi = True

    def update_roi_size(self, position: QPointF) -> None:
        if self.rois:
            self.rois[-1].setSize([position.x() - self.rois[-1].pos().x(), position.y() - self.rois[-1].pos().y()])

    def select_roi(self) -> None:
        self.selected_roi = None
        for roi in self.rois:
            if roi.isUnderMouse():
                self.selected_roi = roi
                break

    def delete_roi(self, roi: pg.ROI) -> None:
        roi.sigClicked.disconnect(self.on_roi_clicked)
        roi.sigRegionChangeStarted.disconnect(self.roi_change_started_signal)
        roi.sigRegionChanged.disconnect(self.roi_changing_signal)
        roi.sigRegionChangeFinished.disconnect(self.roi_change_finished_signal)
        roi.sigClicked.disconnect(self.roi_clicked_signal)

        self.roi_deleted_signal.emit(roi)

        self.viewbox.removeItem(roi)
        self.rois.remove(roi)

    def delete_all_rois(self) -> None:
        for roi in list(self.rois):
            self.delete_roi(roi)

    def delete_selected_roi(self) -> None:
        if self.selected_roi:
            self.delete_roi(self.selected_roi)
            self.selected_roi = None

    def on_roi_clicked(self, roi: pg.ROI) -> None:
        self.selected_roi = roi



import os
from pathlib import Path
from sys import platform


def is_windows() -> bool:
    return "win" in platform


def is_linux() -> bool:
    return "linux" in platform


def is_macos() -> bool:
    return "darwin" in platform


def patch_pathenv_with_default_gpu_paths() -> None:
    if is_windows():
        extra_potential_paths = []
        extra_potential_paths.extend(list(Path(r"C:\Program Files\NVIDIA GPU Computing Toolkit\CUDA").glob("*")))
        extra_potential_paths.extend(list(Path(r"C:\Program Files\NVIDIA GPU Computing Toolkit").glob("*")))
        extra_potential_paths.extend(list(Path(r"C:\Program Files\NVIDIA\CUDNN").glob("*")))
        extra_potential_paths = [str(p) for p in extra_potential_paths]
        separator = ";"
    elif is_linux():
        extra_potential_paths = ["/usr/local/cuda", "/usr/include", "/lib/x86_64-linux-gnu"]
        separator = ":"
    else:
        extra_potential_paths = []
        separator = ":"

    os.environ["PATH"] = separator.join([os.environ["PATH"]] + extra_potential_paths)


patch_pathenv_with_default_gpu_paths()



from typing import Optional, Tuple

import numpy as np
from numpy import fft


def fourier_transform(input_signal: np.ndarray) -> np.ndarray:
    spectrum = fft.fft(input_signal, axis=0)
    spectrum = _remove_negative_frequencies_of_spectrum(spectrum)
    return spectrum


def get_frequencies(len_signal: int, fs: float) -> np.ndarray:
    return fft.rfftfreq(len_signal, 1 / fs)[_get_slice_positive_frequencies(len_signal)]


def inverse_fourier_transform(spectrum: np.ndarray, len_signal: int) -> np.ndarray:
    reconstructed_spectrum = _reconstruct_negative_frequencies_for_spectrum(spectrum, len_signal)
    inverted_signal = fft.ifft(reconstructed_spectrum, n=len_signal, axis=0)
    inverted_signal = np.real(inverted_signal)
    return inverted_signal


def _remove_negative_frequencies_of_spectrum(raw_fft: np.ndarray) -> np.ndarray:
    return raw_fft[_get_slice_positive_frequencies(len(raw_fft))]


def _get_slice_positive_frequencies(len_signal: int) -> slice:
    if len_signal % 2 == 0:
        return slice(0, len_signal // 2 + 1, 1)
    else:
        return slice(0, (len_signal + 1) // 2, 1)


def _reconstruct_negative_frequencies_for_spectrum(spectrum: np.ndarray, len_signal: int) -> np.ndarray:
    if len_signal % 2 == 0:  # Even number of samples
        negative_freqs = np.conj(spectrum[-2:0:-1])  # Exclude DC and Nyquist
    else:  # Odd number of samples
        negative_freqs = np.conj(spectrum[-1:0:-1])  # Exclude only DC
    new_spectrum = np.concatenate((spectrum, negative_freqs))
    return new_spectrum


def get_mean_amplitude_spectrum(data: np.ndarray, fs: float, normalize: bool = True) -> Tuple[np.ndarray, np.ndarray]:
    spec = fourier_transform(data)
    spec = np.abs(spec)
    if data.ndim == 2:
        spec = np.mean(spec, axis=1)
    if normalize:
        spec /= len(data)
    freq = get_frequencies(len(data), fs)
    return freq, spec


def build_amplitude_filter(
    frequencies: np.ndarray,
    f1_f2: Optional[Tuple[float, float]] = None,
    f3_f4: Optional[Tuple[float, float]] = None,
    filter_type: str = "pass",
) -> np.ndarray:
    assert filter_type in ["pass", "reject"]
    assert np.all(frequencies >= 0)
    assert all(
        pair is None or (isinstance(pair, (list, tuple, np.array)) and len(pair) == 2) for pair in [f1_f2, f3_f4]
    )

    if f1_f2 is None and f3_f4 is None:
        if filter_type == "pass":
            return np.ones_like(frequencies)
        else:
            return np.zeros_like(frequencies)

    else:
        amp_filter = np.zeros_like(frequencies)

        if f1_f2:
            f1, f2 = f1_f2
            assert f2 > f1

            rise_part = 0.5 - 0.5 * np.cos(np.pi * (frequencies - f1) / (f2 - f1))
            rise_part[frequencies < f1] = 0
            rise_part[frequencies > f2] = 1

            amp_filter += rise_part

        if f3_f4:
            f3, f4 = f3_f4
            assert f4 > f3

            fall_part = 0.5 + 0.5 * np.cos(np.pi * (frequencies - f3) / (f4 - f3))
            fall_part[frequencies < f3] = 1
            fall_part[frequencies > f4] = 0

            amp_filter += fall_part

        if filter_type == "reject":
            amp_filter = 1 - amp_filter

        return amp_filter


def get_filtered_data(
    data: np.ndarray,
    fs: float,
    f1_f2: Optional[Tuple[float, float]] = None,
    f3_f4: Optional[Tuple[float, float]] = None,
    filter_type: str = "pass",
) -> np.ndarray:
    src_dtype = data.dtype
    length_signal = len(data)
    spectrum = fourier_transform(data)
    frequencies = get_frequencies(length_signal, fs)
    amp_filter = build_amplitude_filter(frequencies, f1_f2, f3_f4, filter_type)
    if data.ndim == 2:
        amp_filter = amp_filter[:, None]
    filtered_spectrum = spectrum * amp_filter
    filtered_data = inverse_fourier_transform(filtered_spectrum, length_signal)
    filtered_data = filtered_data.astype(src_dtype)
    return filtered_data


if __name__ == "__main__":
    t = np.linspace(0, 1)
    y = np.sin(2 * np.pi * t)
    print(get_mean_amplitude_spectrum(y, 1 / (t[1] - t[0])))



from typing import Optional, Tuple

import numpy as np

from first_breaks.data_models.independent import TNormalize
from first_breaks.data_models.initialised_defaults import DEFAULTS
from first_breaks.utils.fourier_transforms import get_filtered_data


def preprocess_gather(
    data: np.ndarray,
    gain: float = DEFAULTS.gain,
    clip: float = DEFAULTS.clip,
    normalize: TNormalize = DEFAULTS.normalize,
    f1_f2: Optional[Tuple[float, float]] = None,
    f3_f4: Optional[Tuple[float, float]] = None,
    fs: Optional[float] = None,
    copy: bool = True,
) -> np.ndarray:
    assert normalize in ["trace", "gather", None]
    assert data.ndim == 2, "Only 2D arrays are acceptable"
    assert isinstance(gain, (int, float, np.number))
    assert isinstance(clip, (int, float, np.number)) and clip >= 0

    if f1_f2 is not None or f3_f4 is not None:
        assert fs

    if copy:
        data = data.copy()

    if normalize is None:
        norma = 1.0
    elif isinstance(normalize, (float, int, np.number)):
        norma = normalize
    elif isinstance(normalize, np.ndarray):
        if normalize.ndim == 1:
            assert (
                len(normalize) == data.shape[1]
            ), "Normalize with numpy have to have the same length as number of traces"
            norma = normalize[None, :]
        elif normalize.ndim == 2:
            assert normalize.shape == (
                1,
                data.shape[1],
            ), "Normalize with numpy have to have the same length as number of traces"
            norma = normalize
        else:
            raise ValueError("Only 1D or 1xD arrays can be used for normalize with numpy")
    elif isinstance(normalize, (list, tuple)):
        assert all(
            isinstance(e, (int, float, np.number)) for e in normalize
        ), "Only numbers can be used for normalize with sequence"
        assert (
            len(normalize) == data.shape[1]
        ), "Normalize with sequence have to have the same length as number of traces"
        norma = np.array(normalize)
    elif normalize == "trace":
        norma = np.nanmean(np.abs(data), axis=0)
        norma[norma < 1e-9 * np.max(norma)] = 1.0  # type: ignore
    elif normalize == "gather":
        norma = np.nanmean(np.abs(data))
        norma = norma if norma > 1e-9 else 1.0
    else:
        raise ValueError("Unsupported `normalize` type and value")

    if f1_f2 is not None or f3_f4 is not None:
        data = get_filtered_data(data=data, f1_f2=f1_f2, f3_f4=f3_f4, fs=fs, filter_type="pass")

    data = data / norma

    data = data * gain
    data = np.clip(data, -clip, clip)
    return data



from math import ceil, floor
from typing import Any, Dict, Optional, Tuple

import pyqtgraph as pg
from PyQt5.QtGui import QCloseEvent, QFont

from first_breaks.desktop.roi_manager import RoiManager, get_rect_of_roi
from first_breaks.sgy.reader import SGY
from first_breaks.utils.fourier_transforms import (
    build_amplitude_filter,
    get_mean_amplitude_spectrum,
)
from first_breaks.utils.utils import resolve_xy2postime


class SpectrumWindow(pg.PlotWidget):
    def __init__(
        self, roi_manager: RoiManager, vsp_view: bool = False, use_open_gl: bool = True, *args: Any, **kwargs: Any
    ):
        super().__init__(useOpenGL=use_open_gl, *args, **kwargs)
        self.setWindowTitle("Amplitude spectrum")

        self.getPlotItem().disableAutoRange()
        self.setAntialiasing(True)
        self.getPlotItem().setClipToView(True)
        self.getPlotItem().setDownsampling(mode="peak")
        self.getPlotItem().showGrid(x=True, y=True, alpha=0.7)

        text_size = 12
        self.label_style = {"font-size": f"{text_size}pt"}
        font = QFont()
        font.setPointSize(text_size)
        self.getPlotItem().getAxis("bottom").setLabel("Frequency, Hz", **self.label_style)
        self.getPlotItem().getAxis("left").setLabel("Amplitude", **self.label_style)

        self.setBackground("w")

        self.roi_manager = roi_manager
        self.roi2line: Dict[pg.ROI, pg.PlotCurveItem] = {}
        self.roi2max_spec: Dict[pg.ROI, float] = {}
        self.roi2max_freq: Dict[pg.ROI, float] = {}
        self.sgy: Optional[SGY] = None
        self.f1_f2: Optional[Tuple[float, float]] = None
        self.f3_f4: Optional[Tuple[float, float]] = None
        self.vsp_view = vsp_view

        self.roi_manager.roi_added_signal.connect(self.add_placeholder_line)
        self.roi_manager.roi_clicked_signal.connect(lambda roi: self.show())
        self.roi_manager.roi_changing_signal.connect(self.update_line)
        self.roi_manager.roi_deleted_signal.connect(self.remove_line)
        self.hide()

    def show_and_highlight_window(self) -> None:
        self.show()
        self.setFocus()
        self.raise_()
        self.activateWindow()

    def update_all(self) -> None:
        for roi in self.roi2line.keys():
            self.update_line(roi)

    def set_sgy(self, sgy: SGY) -> None:
        self.sgy = sgy

    def set_filter_params(self, f1_f2: Optional[Tuple[float, float]], f3_f4: Optional[Tuple[float, float]]) -> None:
        self.f1_f2 = f1_f2
        self.f3_f4 = f3_f4
        self.update_all()

    def set_vsp_view(self, vsp_view: bool) -> None:
        self.vsp_view = vsp_view
        self.update_all()

    def add_placeholder_line(self, roi: pg.ROI) -> None:
        line = pg.PlotCurveItem([], [], pen=roi.currentPen)
        self.roi2line[roi] = line
        self.roi2max_freq[roi] = 0
        self.roi2max_spec[roi] = 0
        self.addItem(line)

    def update_line(self, roi: pg.ROI) -> None:
        x_min, y_min, x_max, y_max = get_rect_of_roi(roi)
        x_min, y_min = resolve_xy2postime(self.vsp_view, x_min, y_min)
        x_max, y_max = resolve_xy2postime(self.vsp_view, x_max, y_max)

        x_min_idx = max(0, ceil(x_min - 1))
        x_max_idx = min(self.sgy.num_traces, floor(x_max - 1))

        y_min_idx = max(0, ceil(self.sgy.ms2index(y_min)))
        y_max_idx = min(self.sgy.num_samples, ceil(self.sgy.ms2index(y_max)))
        if x_max_idx >= x_min_idx and y_max_idx > y_min_idx:
            traces = self.sgy.read_traces_by_ids(
                list(range(x_min_idx, x_max_idx + 1)), min_sample=y_min_idx, max_sample=y_max_idx
            )

            frequencies, spectrum = get_mean_amplitude_spectrum(traces, fs=self.sgy.fs)
            amp_filter = build_amplitude_filter(frequencies, f1_f2=self.f1_f2, f3_f4=self.f3_f4)
            spectrum = amp_filter * spectrum

            self.roi2max_freq[roi] = max(frequencies)
            self.roi2max_spec[roi] = max(spectrum)
        else:
            frequencies, spectrum = [], []
            self.roi2max_freq[roi] = 0
            self.roi2max_spec[roi] = 0

        self.roi2line[roi].setData(frequencies, spectrum)
        self.update_limits()
        self.show_and_highlight_window()

    def remove_line(self, roi: pg.ROI) -> None:
        del self.roi2max_freq[roi]
        del self.roi2max_spec[roi]
        self.removeItem(self.roi2line[roi])
        del self.roi2line[roi]
        self.update_limits()
        self.show_and_highlight_window()

    def update_limits(self) -> None:
        if self.roi2max_freq:
            max_freq = 1.05 * max(list(self.roi2max_freq.values()))
        else:
            max_freq = 1

        if self.roi2max_spec:
            max_spec = 1.1 * max(list(self.roi2max_spec.values()))
        else:
            max_spec = 1

        self.getViewBox().setLimits(xMin=0, xMax=max_freq, yMin=0, yMax=max_spec)
        self.getPlotItem().setYRange(0, max_spec)
        self.getPlotItem().setXRange(0, max_freq)

    def closeEvent(self, e: QCloseEvent) -> None:
        e.ignore()
        self.hide()



from typing import Any, Union

from PyQt5.QtCore import QObject, QRunnable, pyqtBoundSignal, pyqtSignal, pyqtSlot

from first_breaks.picking.ipicker import IPicker
from first_breaks.picking.task import Task


class PickerSignals(QObject):
    started = pyqtSignal()
    progress = pyqtSignal(int)
    finished = pyqtSignal()
    result = pyqtSignal(Task)
    message = pyqtSignal(str)


class PickerQRunnable(QRunnable):
    def __init__(self, picker: IPicker, task: Task, interrpution_signal: Union[pyqtSignal, pyqtBoundSignal]):
        super().__init__()

        self.signals = PickerSignals()

        self.picker = picker
        self.task = task

        self.picker.callback_step_finished = self.callback_step_finished  # type: ignore
        self.picker.callback_processing_started = self.callback_processing_started  # type: ignore
        self.picker.callback_processing_finished = self._do_nothing  # type: ignore

        interrpution_signal.connect(self.picker.callback_interrupt)

        self.len = 0

    def _do_nothing(self, *args: Any, **kwargs: Any) -> None:
        pass

    def callback_step_finished(self, idx_batch: int) -> None:
        progress = int(100 * (idx_batch + 1) / self.len)
        self.signals.progress.emit(progress)

    def callback_processing_started(self, length: int) -> None:  # type: ignore
        self.len = length
        self.signals.progress.emit(0)
        self.signals.message.emit("Picking")

    @pyqtSlot()
    def run(self) -> None:
        self.signals.started.emit()
        self.signals.message.emit("Started")

        try:
            self.task = self.picker.process_task(self.task)
        except Exception as e:
            self.task.success = False
            self.task.error_message = str(e)
            self.task.exception = e
        finally:
            message = "Completed" if self.task.success else "Picking Error"
            self.signals.progress.emit(100)
            self.signals.finished.emit()
            self.signals.message.emit(message)
            self.signals.result.emit(self.task)


class CallInThreadSignals(QObject):
    started = pyqtSignal()
    finished = pyqtSignal()
    success = pyqtSignal(bool)
    result = pyqtSignal(object)


class CallInThread(QRunnable):
    def __init__(self, callable_obj: Any, *args: Any, **kwargs: Any):
        super().__init__()
        self.callable_obj = callable_obj
        self.args = args
        self.kwargs = kwargs
        self.signals = CallInThreadSignals()

    @pyqtSlot()
    def run(self) -> None:
        self.signals.started.emit()
        try:
            result = self.callable_obj(*self.args, **self.kwargs)
            success = True
        except Exception as e:
            result = e
            success = False

        self.signals.finished.emit()
        self.signals.success.emit(success)
        self.signals.result.emit(result)



import os
from os import environ
from pathlib import Path
from sys import platform

from first_breaks import is_linux, is_macos, is_windows


def get_cache_folder() -> Path:
    if is_linux():
        return Path(environ.get("XDG_CACHE_HOME", Path.home() / ".cache")) / "first_breaks_picking"
    elif is_macos():
        return Path.home() / "Library" / "Caches" / "first_breaks_picking"
    elif is_windows():
        return Path.home() / ".cache" / "first_breaks_picking"
    else:
        raise ValueError(f"Unexpected platform {platform}.")


PROJECT_ROOT = Path(__file__).parent.parent
CACHE_FOLDER = get_cache_folder()

DEMO_SGY_PATH = CACHE_FOLDER / "real_gather.sgy"
DEMO_SGY_URL = "https://raw.githubusercontent.com/DaloroAT/first_breaks_picking/main/data/real_gather.sgy"
DEMO_SGY_HASH = "92fe2992b57d69c6f572c672f63960cf"


MODEL_ONNX_PATH = CACHE_FOLDER / "fb.onnx"
MODEL_ONNX_URL = "https://oml.daloroserver.com/download/seis/fb.onnx"
MODEL_ONNX_HASH = "7e39e017b01325180e36885eccaeb17a"

TIMEOUT = 60

HIGH_DPI = os.getenv("HIGH_DPI", True)



import sys
from typing import Any, Callable, Optional, Sequence, Tuple, Union
from weakref import WeakKeyDictionary

from PyQt5.QtCore import (
    QEasingCurve,
    QObject,
    QPoint,
    QPropertyAnimation,
    Qt,
    QTimer,
    pyqtBoundSignal,
    pyqtSignal,
)
from PyQt5.QtGui import QColor, QPalette
from PyQt5.QtWidgets import (
    QApplication,
    QFrame,
    QLabel,
    QLineEdit,
    QVBoxLayout,
    QWidget,
)

TDoUntil = Union[int, float, pyqtSignal, Sequence[pyqtSignal]]


def run_do_until(do_until: TDoUntil, func: Callable[[Any], Any]) -> None:
    if isinstance(do_until, (int, float)):
        timeout = max(1, int(do_until))
        QTimer.singleShot(timeout, func)
        return
    elif isinstance(do_until, (pyqtSignal, pyqtBoundSignal)):
        do_until = [do_until]
    for signal in do_until:
        signal.connect(func)


class TextToolTip(QLabel):
    def __init__(  # type: ignore
        self,
        widget: QWidget,
        text: str,
        text_position="right",
        color: Union[Tuple[int, int, int], str] = "#faf3be",
        do_until: TDoUntil = 5000,
    ):
        super().__init__(widget)

        self.setWordWrap(True)
        self.setFrameShape(QFrame.StyledPanel)
        self.setStyleSheet(
            f"""
                    background-color: {color};
                    border: 1px solid black;
                    padding: 2px;
                """
        )
        self.setAutoFillBackground(True)
        self.setWindowFlags(Qt.ToolTip | Qt.WindowStaysOnTopHint)

        self.setText(text)
        self.adjustSize()

        global_pos = widget.mapToGlobal(QPoint(0, 0))
        if text_position == "top":
            self.move(global_pos.x(), global_pos.y() - self.height() - 5)
        elif text_position == "bottom":
            self.move(global_pos.x(), global_pos.y() + widget.height() + 5)
        elif text_position == "left":
            self.move(global_pos.x() - self.width() - 5, global_pos.y())
        elif text_position == "right":
            self.move(global_pos.x() + widget.width() + 5, global_pos.y())
        else:
            raise ValueError("Wrong position value")
        self.show()

        run_do_until(do_until, self.close)  # type: ignore


class _HighlightManager:
    _original_colors = WeakKeyDictionary()  # type: ignore
    _highlight_counts = WeakKeyDictionary()  # type: ignore

    @classmethod
    def highlight_widget(cls, widget: QWidget, color: Tuple[int, int, int]) -> None:
        if widget not in cls._original_colors:
            cls._original_colors[widget] = widget.palette()

        if widget not in cls._highlight_counts:
            cls._highlight_counts[widget] = 0

        cls._highlight_counts[widget] += 1
        palette = widget.palette()
        palette.setColor(QPalette.Base, QColor(*color))
        widget.setPalette(palette)

    @classmethod
    def remove_highlight(cls, widget: QWidget) -> None:
        if widget in cls._highlight_counts:
            cls._highlight_counts[widget] -= 1
            if cls._highlight_counts[widget] <= 0:
                widget.setPalette(cls._original_colors[widget])
                del cls._original_colors[widget]
                del cls._highlight_counts[widget]


class HighlightToolTip(QObject):
    _manager = _HighlightManager()

    def __init__(  # type: ignore
        self,
        widgets: Union[QWidget, Sequence[QWidget]],
        parent: Optional[QWidget] = None,
        do_until: TDoUntil = 5000,
        color=(255, 220, 220),
    ):
        if not parent and widgets:
            if isinstance(widgets, (list, tuple)):
                parent = widgets[0]
            else:
                parent = widgets
        super().__init__(parent)

        self.widgets = widgets if isinstance(widgets, (list, tuple)) else [widgets]

        for widget in self.widgets:
            self._manager.highlight_widget(widget, color)

        run_do_until(do_until, self.remove_highlights)  # type: ignore

    def remove_highlights(self) -> None:
        for widget in self.widgets:
            self._manager.remove_highlight(widget)
        self.deleteLater()


class _Shaker:
    def __init__(self, widget: QWidget):
        self._widget = widget
        self._start_pos = widget.pos()
        self._animation = QPropertyAnimation(widget, b"pos")
        self._animation.setStartValue(self._start_pos)
        self._animation.setEndValue(self._start_pos)

        basic_duration = 1000
        shakes_per_second = 10

        num_shakes = max(1, int(shakes_per_second * basic_duration / 1000))
        for i in range(num_shakes):
            self._animation.setKeyValueAt((i + 1) / num_shakes, self._start_pos + (-1) ** i * QPoint(2, 0))

        self._animation.setDuration(basic_duration)
        self._animation.setLoopCount(-1)
        self._animation.setEasingCurve(QEasingCurve.InOutQuad)

    def shake(self) -> None:
        self._animation.start()

    def stop(self) -> None:
        self._animation.stop()
        self._widget.move(self._start_pos)


class ShakeToolTip(QObject):
    def __init__(
        self, widgets: Union[QWidget, Sequence[QWidget]], parent: Optional[QWidget] = None, do_until: TDoUntil = 5000
    ):
        if not parent and widgets:
            if isinstance(widgets, (list, tuple)):
                parent = widgets[0]
            else:
                parent = widgets
        super().__init__(parent)

        self._shakers = []
        self.widgets = widgets if isinstance(widgets, (list, tuple)) else [widgets]

        for widget in self.widgets:
            shaker = _Shaker(widget)
            shaker.shake()
            self._shakers.append(shaker)

        run_do_until(do_until, self.stop_shaking)  # type: ignore

    def stop_shaking(self) -> None:
        for shaker in self._shakers:
            shaker.stop()
        self.deleteLater()


if __name__ == "__main__":

    class ValidatorWidget(QWidget):
        def __init__(self):  # type: ignore
            super().__init__()

            layout = QVBoxLayout(self)
            self.line_edit = QLineEdit(self)
            layout.addWidget(self.line_edit)

            # Signal connection
            self.line_edit.textChanged.connect(self.validate_input)

        def validate_input(self, text: str) -> None:
            if "@" not in text:
                TextToolTip(
                    self.line_edit, text="Password must contain '@' symbol!" * 1, do_until=self.line_edit.textChanged
                )
                HighlightToolTip(self.line_edit, do_until=self.line_edit.textChanged)
                ShakeToolTip(self.line_edit, do_until=self.line_edit.textChanged)

    app = QApplication(sys.argv)
    window = ValidatorWidget()
    window.show()
    sys.exit(app.exec_())



import ast
import os
import warnings
from pathlib import Path
from typing import Any, List, Optional, Sequence, Tuple, Union

import numpy as np
import pyqtgraph as pg
from PyQt5.QtCore import Qt, QTimer, pyqtSlot
from PyQt5.QtGui import QCloseEvent, QColor, QFont, QPainterPath, QPen
from PyQt5.QtWidgets import QApplication
from pyqtgraph import AxisItem
from pyqtgraph.exporters import ImageExporter
from pyqtgraph.GraphicsScene.mouseEvents import MouseClickEvent

from first_breaks.const import HIGH_DPI
from first_breaks.data_models.independent import TColor, TNormalize
from first_breaks.data_models.initialised_defaults import DEFAULTS
from first_breaks.desktop.roi_manager import RoiManager
from first_breaks.desktop.spectrum_window import SpectrumWindow
from first_breaks.picking.task import Task
from first_breaks.picking.utils import preprocess_gather
from first_breaks.sgy.reader import SGY
from first_breaks.utils.utils import resolve_postime2xy as postime2xy
from first_breaks.utils.utils import resolve_xy2postime as xy2postime

if HIGH_DPI:
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)


class GraphWidget(pg.PlotWidget):
    def __init__(self, use_open_gl: bool = True, *args: Any, **kwargs: Any):
        super().__init__(useOpenGL=use_open_gl, *args, **kwargs)
        self.plotItem.disableAutoRange()
        self.setAntialiasing(False)
        self.plotItem.setClipToView(True)
        self.plotItem.setDownsampling(mode="peak")
        self.plotItem.ctrlMenu = None

        self.invert_x = DEFAULTS.invert_x
        self.invert_y = DEFAULTS.invert_y
        self.vsp_view = DEFAULTS.vsp_view
        self.sgy: Optional[SGY] = None
        self.nn_picks_in_ms: Optional[np.ndarray] = None
        self.nn_picks_as_item: Optional[pg.PlotCurveItem] = None
        self.extra_picks_as_item_list: Optional[List[pg.PlotCurveItem]] = []
        self.processing_region_as_items: List[pg.QtWidgets.QGraphicsPathItem] = []
        self.traces_as_items: List[pg.QtWidgets.QGraphicsPathItem] = []
        self.is_picks_modified_manually = False
        self.pos_ax_header: Optional[str] = None

        self.x_ax: Optional[AxisItem] = None
        self.y_ax: Optional[AxisItem] = None
        self.pos_ax: Optional[AxisItem] = None
        self.time_ax: Optional[AxisItem] = None
        self.setup_axes()

        self.spectrum_roi_manager = RoiManager(viewbox=self.getViewBox())
        self.spectrum_window = SpectrumWindow(use_open_gl=use_open_gl, roi_manager=self.spectrum_roi_manager)
        self.mouse_click_signal = pg.SignalProxy(self.sceneObj.sigMouseClicked, rateLimit=60, slot=self.mouse_clicked)

    def resolve_postime2xy(self, position: Any, time: Any) -> Tuple[Any, Any]:
        return postime2xy(vsp_view=self.vsp_view, position=position, time=time)

    def resolve_xy2postime(self, x: Any, y: Any) -> Tuple[Any, Any]:
        return xy2postime(vsp_view=self.vsp_view, x=x, y=y)

    def setup_axes(self) -> None:
        self.getPlotItem().invertX(self.invert_x)
        self.getPlotItem().invertY(self.invert_y)
        self.getPlotItem().showAxis("top", True)
        self.getPlotItem().showAxis("bottom", False)

        self.x_ax = self.getPlotItem().getAxis("top")
        self.y_ax = self.getPlotItem().getAxis("left")

        self.pos_ax, self.time_ax = self.resolve_xy2postime(self.x_ax, self.y_ax)

        self.pos_ax.tickStrings = self.replace_tick_labels

        text_size = 12
        self.label_style = {"font-size": f"{text_size}pt"}
        font = QFont()
        font.setPointSize(text_size)
        self.pos_ax.setLabel(None, **self.label_style)
        self.time_ax.setLabel("t, ms", **self.label_style)
        self.pos_ax.setTickFont(font)
        self.time_ax.setTickFont(font)

    def full_clean(self) -> None:
        self.remove_nn_picks()
        self.remove_traces()
        self.remove_processing_region()
        self.nn_picks_as_item = None
        self.nn_picks_in_ms = None
        self.is_picks_modified_manually = False
        self.spectrum_roi_manager.delete_all_rois()
        self.clear()

    def plotseis(
        self,
        sgy: SGY,
        clip: float = DEFAULTS.clip,
        gain: float = DEFAULTS.gain,
        normalize: TNormalize = DEFAULTS.normalize,
        f1_f2: Optional[Tuple[float, float]] = None,
        f3_f4: Optional[Tuple[float, float]] = None,
        x_axis: Optional[str] = DEFAULTS.x_axis,
        fill_black: Optional[str] = DEFAULTS.fill_black,
        vsp_view: bool = DEFAULTS.vsp_view,
        refresh_view: bool = True,
        invert_x: bool = DEFAULTS.invert_x,
        invert_y: bool = DEFAULTS.invert_y,
    ) -> None:
        self.pos_ax_header = x_axis

        self.sgy = sgy
        self.spectrum_window.set_sgy(sgy)
        self.spectrum_window.set_filter_params(f1_f2, f3_f4)

        traces = self.sgy.read()

        traces = preprocess_gather(
            traces, gain=gain, clip=clip, normalize=normalize, f1_f2=f1_f2, f3_f4=f3_f4, fs=self.sgy.fs, copy=True
        )

        # we put clearing after preprocessing to reduce time when user see nothing
        self.clear()
        # axes related checks
        if self.vsp_view != vsp_view:
            self.vsp_view = vsp_view
            self.spectrum_roi_manager.change_rois_view()
            self.spectrum_window.set_vsp_view(self.vsp_view)
            need_to_refresh = True
        else:
            need_to_refresh = False
        if self.invert_x != invert_x:
            self.invert_x = invert_x
            need_to_refresh |= True
        else:
            need_to_refresh |= False
        if self.invert_y != invert_y:
            self.invert_y = invert_y
            need_to_refresh |= True
        else:
            need_to_refresh |= False

        if need_to_refresh:
            self.setup_axes()

        num_sample, num_traces = self.sgy.shape
        t = np.arange(num_sample) * self.sgy.dt_ms

        pos_max = num_traces + 1
        time_max = t[-1]
        x_max, y_max = self.resolve_postime2xy(pos_max, time_max)

        self.getViewBox().setLimits(xMin=0, xMax=x_max, yMin=0, yMax=y_max)

        if refresh_view or need_to_refresh:
            self.getPlotItem().setYRange(min=0, max=y_max)
            self.getPlotItem().setXRange(min=0, max=x_max)

        for idx in range(num_traces):
            self._plot_trace_fast(trace=traces[:, idx], time=t, shift=idx + 1, fill_black=fill_black)

        self.pos_ax.showLabel()

    def _plot_trace_fast(self, trace: np.ndarray, time: np.ndarray, shift: int, fill_black: Optional[str]) -> None:
        connect = np.ones(len(time), dtype=np.int32)
        connect[-1] = 0

        trace[0] = 0
        trace[-1] = 0

        shifted_trace = trace + shift

        path_x, path_y = self.resolve_postime2xy(shifted_trace, time)
        path = pg.arrayToQPath(x=path_x, y=path_y, connect=connect)

        item = pg.QtWidgets.QGraphicsPathItem(path)
        pen = QPen(Qt.black, 1, Qt.SolidLine, Qt.FlatCap, Qt.MiterJoin)
        pen.setWidth(0.1)
        item.setPen(pen)
        item.setBrush(Qt.white)
        self.addItem(item)
        self.traces_as_items.append(item)

        if fill_black is None:
            return
        else:
            sign = -1 if fill_black == "left" else 1

            x, y = self.resolve_postime2xy(shift, time[0])
            w, h = self.resolve_postime2xy(sign * 1.1 * max(np.abs(shifted_trace)), time[-1])

            rect = QPainterPath()
            rect.addRect(x, y, w, h)

            patch = path.intersected(rect)
            item = pg.QtWidgets.QGraphicsPathItem(patch)

            pen = QPen(QColor(255, 255, 255, 0), 1, Qt.SolidLine, Qt.FlatCap, Qt.MiterJoin)
            pen.setWidth(0.1)
            item.setPen(pen)
            item.setBrush(Qt.black)
            self.addItem(item)
            self.traces_as_items.append(item)

    def replace_tick_labels(self, *args: Any, **kwargs: Any) -> List[str]:
        self.pos_ax.setLabel(self.pos_ax_header, **self.label_style)
        previous_labels = AxisItem.tickStrings(self.pos_ax, *args, **kwargs)

        if self.pos_ax_header is not None:
            labels_from_headers = []
            for v in previous_labels:
                v = ast.literal_eval(v)
                if v % 1 == 0:
                    v = int(v) - 1
                    if 0 <= v < self.sgy.num_traces:
                        labels_from_headers.append(str(self.sgy.traces_headers[self.pos_ax_header].iloc[v]))
                    else:
                        labels_from_headers.append("")
                else:
                    labels_from_headers.append("")
            return labels_from_headers
        else:
            return previous_labels

    def remove_nn_picks(self) -> None:
        self.is_picks_modified_manually = False
        if self.nn_picks_as_item:
            self.removeItem(self.nn_picks_as_item)
            self.nn_picks_as_item = None
            self.nn_picks_in_ms = None

    def remove_extra_picks(self) -> None:
        for item in self.extra_picks_as_item_list:
            self.removeItem(item)

    def remove_processing_region(self) -> None:
        if self.processing_region_as_items:
            for item in self.processing_region_as_items:
                self.removeItem(item)

    def remove_traces(self) -> None:
        if self.traces_as_items:
            for item in self.traces_as_items:
                self.removeItem(item)

    def plot_processing_region(
        self,
        traces_per_gather: int,
        maximum_time: float,
        region_contour_color: TColor = DEFAULTS.region_contour_color,
        region_poly_color: TColor = DEFAULTS.region_poly_color,
        region_contour_width: float = DEFAULTS.region_contour_width,
    ) -> None:
        self.remove_processing_region()

        num_sample, num_traces = self.sgy.shape
        sgy_end_time = (num_sample + 2) * self.sgy.dt_ms
        region_start_time = maximum_time if maximum_time > 0 else sgy_end_time

        contour_pen = QPen(QColor(*region_contour_color), region_contour_width, Qt.DashLine, Qt.FlatCap, Qt.MiterJoin)
        poly_brush = QColor(*region_poly_color)

        # Vertical lines
        line_time = np.array([0, region_start_time])
        for idx in np.arange(traces_per_gather + 0.5, num_traces - 1, traces_per_gather):
            line_pos = np.array([idx, idx])
            line_x, line_y = self.resolve_postime2xy(line_pos, line_time)
            line_path = pg.arrayToQPath(line_x, line_y, np.ones(2, dtype=np.int32))
            line_item = pg.QtWidgets.QGraphicsPathItem(line_path)
            line_item.setPen(contour_pen)
            self.processing_region_as_items.append(line_item)
            self.addItem(line_item)

        # Transparent polygon on ignored part
        poly_pos = np.array([-2, num_traces + 2, num_traces + 2, -2])
        poly_time = np.array([region_start_time, region_start_time, sgy_end_time, sgy_end_time])
        poly_x, poly_y = self.resolve_postime2xy(poly_pos, poly_time)
        poly_path = pg.arrayToQPath(poly_x, poly_y, np.ones(4, dtype=np.int32))
        poly_item = pg.QtWidgets.QGraphicsPathItem(poly_path)
        poly_item.setPen(contour_pen)
        poly_item.setBrush(poly_brush)
        self.processing_region_as_items.append(poly_item)
        self.addItem(poly_item)

    def get_picks_as_item(
        self,
        picks_ms: Sequence[float],
        color: TColor = DEFAULTS.picks_color,
    ) -> pg.PlotCurveItem:
        x, y = self.resolve_postime2xy(np.arange(self.sgy.num_traces) + 1, np.array(picks_ms))

        line = pg.PlotCurveItem()
        line.setData(x, y)
        pen = pg.mkPen(color=color, width=3)
        line.setPen(pen)

        return line

    def plot_nn_picks(self, picks_ms: Sequence[float], color: TColor = DEFAULTS.picks_color) -> None:
        self.remove_nn_picks()
        self.is_picks_modified_manually = False

        self.nn_picks_in_ms = np.array(picks_ms)
        self.nn_picks_as_item = self.get_picks_as_item(self.nn_picks_in_ms, color)

        self.addItem(self.nn_picks_as_item)

    def plot_extra_picks(self, picks_ms: Sequence[float], color: TColor = DEFAULTS.picks_color) -> None:
        picks = self.get_picks_as_item(picks_ms, color)

        self.addItem(picks)
        self.extra_picks_as_item_list.append(picks)

    def mouse_clicked(self, ev: Tuple[MouseClickEvent]) -> None:
        ev = ev[0]
        if self.nn_picks_as_item is not None and ev.button() == 1:
            mouse_xy = self.getPlotItem().vb.mapSceneToView(ev.scenePos())
            mouse_pos, mouse_time = self.resolve_xy2postime(mouse_xy.x(), mouse_xy.y())

            picks_x, picks_y = self.nn_picks_as_item.getData()
            picks_pos, picks_time = self.resolve_xy2postime(picks_x, picks_y)

            closest = np.argmin(np.abs(picks_pos - mouse_pos))
            mouse_time = np.clip(mouse_time, 0, self.sgy.max_time_ms)
            picks_time[closest] = mouse_time

            picks_x, picks_y = self.resolve_postime2xy(picks_pos, picks_time)
            self.nn_picks_as_item.setData(picks_x, picks_y)
            self.nn_picks_in_ms = picks_time
            self.is_picks_modified_manually = True

    def closeEvent(self, e: QCloseEvent) -> None:
        self.spectrum_window.close()
        e.accept()


class HighMemoryConsumption(Exception):
    pass


class UnsupportedImageSize(Exception):
    pass


need_kwargs_exception = ValueError(
    "Please, use named arguments instead of ordered for visualizations parameters. "
    "E.g. instead of `export(sgy, 'gather.png', 1.5)` use"
    " `export(sgy, 'gather.png', clip=1.5)`"
)


class GraphExporter(GraphWidget):
    MAX_SIDE_SIZE = 65000
    MAX_NUM_PIXELS = MAX_SIDE_SIZE * 2000

    def __init__(self, *args: Any, **kwargs: Any):
        os.environ["DEBIAN_FRONTEND"] = "noninteractive"
        os.environ["LIBGL_ALWAYS_INDIRECT"] = "1"
        os.environ["QT_QPA_PLATFORM"] = "offscreen"
        super().__init__(*args, **kwargs)
        self.setAntialiasing(True)

    @classmethod
    def avoid_memory_bomb(
        cls,
        height_image: int,
        width_image: int,
    ) -> None:

        if height_image > cls.MAX_SIDE_SIZE:
            message = (
                f"It is not possible to render a picture of the given height = {height_image}. "
                f"Max height = {cls.MAX_SIDE_SIZE}. Decrease rendering parameters."
            )
            raise UnsupportedImageSize(message)

        if width_image > cls.MAX_SIDE_SIZE:
            message = (
                f"It is not possible to render a picture of the given width = {width_image}. "
                f"Max width = {cls.MAX_SIDE_SIZE}. Decrease rendering parameters."
            )
            raise UnsupportedImageSize(message)

        num_pixels = height_image * width_image

        if num_pixels > cls.MAX_NUM_PIXELS:
            message = (
                f"The size of the picture will turn out to be too large ({num_pixels} pixels) for "
                f"this SGY file. Decrease rendering parameters."
            )
            raise HighMemoryConsumption(message)

    def export(
        self,
        sgy: SGY,
        image_filename: Optional[Union[str, Path]],
        *args: Any,
        # content parameters
        clip: float = DEFAULTS.clip,
        gain: float = DEFAULTS.gain,
        normalize: TNormalize = DEFAULTS.normalize,
        fill_black: Optional[str] = DEFAULTS.fill_black,
        time_window: Optional[Tuple[float, float]] = None,
        traces_window: Optional[Tuple[float, float]] = None,
        picks_ms: Optional[Sequence[float]] = None,
        task: Optional[Task] = None,
        show_processing_region: bool = True,
        picks_color: TColor = DEFAULTS.picks_color,
        contour_color: TColor = DEFAULTS.region_contour_color,
        poly_color: TColor = DEFAULTS.region_poly_color,
        contour_width: float = DEFAULTS.region_contour_width,
        x_axis: Optional[str] = DEFAULTS.x_axis,
        # rendering parameters
        height: int = 500,
        width: Optional[int] = None,
        width_per_trace: int = 20,
        headers_total_pixels: int = 50,
        headers_font_pixels: Optional[int] = None,
        time_spacing: Optional[int] = None,
        traces_spacing: Optional[int] = None,
        hide_traces_axis: bool = False,
    ) -> None:
        if args:
            raise need_kwargs_exception

        if picks_ms is not None and task is not None:
            raise ValueError("'picks_ms' and 'task' are mutually exclusive. Use only one of them or none")

        if width is None:
            if traces_window is None:
                num_traces = sgy.num_traces
                width = int(width_per_trace * num_traces) + headers_total_pixels
            else:
                width = int(width_per_trace * (traces_window[1] - traces_window[0])) + headers_total_pixels

        self.avoid_memory_bomb(height, width)

        self.x_ax_header = x_axis

        self.clear()
        self.plotseis(sgy, normalize=normalize, clip=clip, gain=gain, fill_black=fill_black, refresh_view=True)

        if task:
            picks_to_plot = task.picks_in_ms  # type: ignore
        elif picks_ms is not None:
            picks_to_plot = picks_ms  # type: ignore
        else:
            picks_to_plot = None  # type: ignore

        if picks_to_plot is not None:
            self.plot_nn_picks(picks_to_plot, color=picks_color)

        if task is not None and show_processing_region:
            self.plot_processing_region(
                maximum_time=task.maximum_time,
                traces_per_gather=task.traces_per_gather,
                region_contour_color=contour_color,
                region_poly_color=poly_color,
                region_contour_width=contour_width,
            )

        if time_window:
            self.getPlotItem().setYRange(time_window[0], time_window[1], padding=0)
        if traces_window:
            self.getPlotItem().setXRange(traces_window[0], traces_window[1], padding=0)

        headers_font_pixels = headers_font_pixels or int(0.35 * headers_total_pixels)
        labelstyle = {"font-size": f"{headers_font_pixels}px"}
        tickfont = QFont()
        tickfont.setPixelSize(max(int(0.9 * headers_font_pixels), 1))
        self.x_ax.setLabel(self.x_ax_header, **labelstyle)
        self.y_ax.setLabel("t, ms", **labelstyle)
        self.x_ax.setTickFont(tickfont)
        self.y_ax.setTickFont(tickfont)

        self.x_ax.setHeight(headers_total_pixels)

        if time_spacing:
            self.y_ax.setTickSpacing(time_spacing, time_spacing)
        if traces_spacing:
            self.x_ax.setTickSpacing(traces_spacing, time_spacing)

        self.y_ax.setWidth(headers_total_pixels)
        if hide_traces_axis:
            self.x_ax.showLabel(False)
            self.x_ax.setTicks([])
            self.x_ax.setHeight(0)
        else:
            self.x_ax.setHeight(headers_total_pixels)

        self.plotItem.setFixedHeight(height)
        self.plotItem.setFixedWidth(width)

        image = ImageExporter(self.plotItem).export(toBytes=True)

        if image_filename:
            Path(image_filename).parent.mkdir(exist_ok=True, parents=True)
            image.save(str(image_filename), quality=100)

        QTimer.singleShot(0, self.close_widget)

    @pyqtSlot()
    def close_widget(self) -> None:
        self.close()


def export_image(
    source: Union[str, Path, bytes, np.ndarray, SGY, Task],
    image_filename: Optional[Union[str, Path]],
    *args: Any,
    dt_mcs: float = 1e3,
    clip: float = DEFAULTS.clip,
    gain: float = DEFAULTS.gain,
    normalize: TNormalize = DEFAULTS.normalize,
    fill_black: Optional[str] = DEFAULTS.fill_black,
    time_window: Optional[Tuple[float, float]] = None,
    traces_window: Optional[Tuple[float, float]] = None,
    picks_ms: Optional[Sequence[float]] = None,
    show_processing_region: bool = True,
    picks_color: TColor = DEFAULTS.picks_color,
    contour_color: TColor = DEFAULTS.region_contour_color,
    poly_color: TColor = DEFAULTS.region_poly_color,
    contour_width: float = DEFAULTS.region_contour_width,
    x_axis: Optional[str] = DEFAULTS.x_axis,
    # rendering parameters
    height: int = 500,
    width: Optional[int] = None,
    width_per_trace: int = 20,
    headers_total_pixels: int = 50,
    headers_font_pixels: Optional[int] = None,
    time_spacing: Optional[int] = None,
    traces_spacing: Optional[int] = None,
    hide_traces_axis: bool = False,
) -> None:
    if args:
        raise need_kwargs_exception

    if isinstance(source, (str, Path, bytes)):
        sgy = SGY(source)
        task = None
    elif isinstance(source, np.ndarray):
        sgy = SGY(source, dt_mcs=dt_mcs)
        task = None
    elif isinstance(source, SGY):
        sgy = source
        task = None
    elif isinstance(source, Task):
        sgy = source.sgy
        task = source
    else:
        raise TypeError("Unsupported type for 'source'")

    warnings.filterwarnings("ignore")
    app = QApplication([])
    app.setQuitOnLastWindowClosed(True)
    window = GraphExporter(background="w")
    window.hide()
    window.setAntialiasing(True)
    window.export(
        sgy=sgy,
        image_filename=image_filename,
        clip=clip,
        gain=gain,
        normalize=normalize,
        fill_black=fill_black,
        time_window=time_window,
        traces_window=traces_window,
        picks_ms=picks_ms,
        task=task,
        show_processing_region=show_processing_region,
        picks_color=picks_color,
        contour_color=contour_color,
        poly_color=poly_color,
        contour_width=contour_width,
        x_axis=x_axis,
        height=height,
        width=width,
        width_per_trace=width_per_trace,
        headers_total_pixels=headers_total_pixels,
        headers_font_pixels=headers_font_pixels,
        time_spacing=time_spacing,
        traces_spacing=traces_spacing,
        hide_traces_axis=hide_traces_axis,
    )
    app.exec()
    warnings.resetwarnings()


if __name__ == "__main__":
    from first_breaks.sgy.reader import SGY
    from first_breaks.utils.utils import download_demo_sgy

    demo_sgy = download_demo_sgy()
    export_image(demo_sgy, "demo_sgy.png")
    # app = QApplication([])
    # window = GraphWidget(background="w")
    # window.show()
    # window.plotseis(SGY(demo_sgy))
    # app.exec_()



import sys
import warnings
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union

from PyQt5.QtCore import QSize, Qt, QThreadPool
from PyQt5.QtGui import QCloseEvent
from PyQt5.QtWidgets import (
    QAction,
    QApplication,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QProgressBar,
    QSizePolicy,
    QStyle,
    QToolBar,
    QWidget,
)

from first_breaks.const import DEMO_SGY_PATH, HIGH_DPI, MODEL_ONNX_HASH, MODEL_ONNX_PATH
from first_breaks.data_models.dependent import TraceHeaderParams
from first_breaks.data_models.independent import ExceptionOptional
from first_breaks.desktop.byte_encode_unit_widget import QDialogByteEncodeUnit
from first_breaks.desktop.graph import GraphWidget
from first_breaks.desktop.nn_manager import NNManager
from first_breaks.desktop.settings_processing_widget import (
    PickingSettings,
    PicksFromFileSettings,
    PlotseisSettings,
    SettingsProcessingWidget,
)
from first_breaks.desktop.utils import MessageBox, set_geometry
from first_breaks.picking.task import Task
from first_breaks.sgy.reader import SGY
from first_breaks.utils.utils import (
    UnitsConverter,
    calc_hash,
    download_demo_sgy,
    download_model_onnx,
    multiply_iterable_by,
)

warnings.filterwarnings("ignore")

if HIGH_DPI:
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)


class FileState:
    valid_file = 0
    file_not_exists = 1
    file_changed = 2

    @classmethod
    def get_file_state(cls, fname: Union[str, Path], fhash: str) -> int:
        if not Path(fname).is_file():
            return cls.file_not_exists
        else:
            return cls.valid_file if calc_hash(fname) == fhash else cls.file_changed


class ReadyToProcess:
    sgy_selected: bool = False
    model_loaded: bool = False

    def is_ready(self) -> bool:
        return (self.sgy_selected == self.model_loaded) is True


class MainWindow(QMainWindow):
    def __init__(self, use_open_gl: bool = True):  # type: ignore
        super(MainWindow, self).__init__()

        if getattr(sys, "frozen", False):
            self.main_folder = Path(sys._MEIPASS)  # type: ignore
        else:
            self.main_folder = Path(__file__).parent

        set_geometry(self, width_rel=0.6, height_rel=0.6, fix_size=False, centralize=True)
        self.setWindowTitle("First breaks picking")

        self.threadpool = QThreadPool()

        # toolbar
        self.toolbar = QToolBar()
        self.toolbar.setIconSize(QSize(30, 30))
        self.addToolBar(self.toolbar)

        # buttons on toolbar
        icon_load_nn = self.style().standardIcon(QStyle.SP_ComputerIcon)
        # icon_load_nn = QIcon(str(self.main_folder / "icons" / "nn.png"))
        self.button_load_nn = QAction(icon_load_nn, "Load model", self)
        self.button_load_nn.triggered.connect(self.load_nn)
        self.button_load_nn.setEnabled(True)
        self.toolbar.addAction(self.button_load_nn)

        icon_get_filename = self.style().standardIcon(QStyle.SP_DirIcon)
        # icon_get_filename = QIcon(str(self.main_folder / "icons" / "sgy.png"))
        self.button_get_filename = QAction(icon_get_filename, "Open SGY-file", self)
        self.button_get_filename.triggered.connect(self.get_filename)
        self.button_get_filename.setEnabled(True)
        self.toolbar.addAction(self.button_get_filename)

        self.toolbar.addSeparator()

        icon_visual_settings = self.style().standardIcon(QStyle.SP_FileDialogContentsView)
        self.button_settings_processing = QAction(icon_visual_settings, "Settings and Processing", self)
        self.button_settings_processing.triggered.connect(self.show_settings_processing_window)
        self.button_settings_processing.setEnabled(False)
        self.toolbar.addAction(self.button_settings_processing)

        self.need_processing_region = True
        icon_processing_show = self.style().standardIcon(QStyle.SP_FileDialogListView)
        # icon_export = QIcon(str(self.main_folder / "icons" / "export.png"))
        self.button_processing_show = QAction(icon_processing_show, "Show processing grid", self)
        self.button_processing_show.triggered.connect(self.processing_region_changed)
        self.button_processing_show.setChecked(self.need_processing_region)
        self.button_processing_show.setEnabled(True)
        self.button_processing_show.setCheckable(True)
        if self.need_processing_region:
            self.button_processing_show.toggle()
        self.toolbar.addAction(self.button_processing_show)

        self.toolbar.addSeparator()

        icon_export = self.style().standardIcon(QStyle.SP_DialogSaveButton)
        # icon_export = QIcon(str(self.main_folder / "icons" / "export.png"))
        self.button_export = QAction(icon_export, "Export picks to file", self)
        self.button_export.triggered.connect(self.export)
        self.button_export.setEnabled(False)
        self.toolbar.addAction(self.button_export)

        spacer = QWidget(self)
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.toolbar.addWidget(spacer)

        self.status = self.statusBar()
        self.status_progress = QProgressBar()
        self.status_progress.hide()

        self.status_message = QLabel()
        self.status_message.setText("Open SGY file or load model")

        status_widget = QWidget()
        status_layout = QHBoxLayout()
        status_widget.setLayout(status_layout)
        status_layout.addWidget(self.status_progress)
        status_layout.addWidget(self.status_message)

        self.status.addPermanentWidget(status_widget)

        # graph widget
        self.graph = GraphWidget(use_open_gl=use_open_gl, background="w")
        self.graph.hide()
        self.setCentralWidget(self.graph)

        # visual settings widget
        self.plotseis_settings = PlotseisSettings()
        first_byte = 1
        self.picks_from_file_settings = PicksFromFileSettings(byte_position=first_byte)
        self.settings_processing_widget = SettingsProcessingWidget(
            hide_on_close=True,
            first_byte=first_byte,
            **{**self.plotseis_settings.model_dump(), **self.picks_from_file_settings.model_dump()},
        )
        self.settings_processing_widget.hide()
        self.settings_processing_widget.export_plotseis_settings_signal.connect(self.update_plotseis_settings)
        self.settings_processing_widget.export_picking_settings_signal.connect(self.pick_fb)
        self.settings_processing_widget.export_picks_from_file_settings_signal.connect(
            self.update_picks_from_file_settings
        )
        self.settings_processing_widget.toggle_picks_from_file_signal.connect(self.toggle_picks_from_file)

        # nn manager
        self.nn_manager = NNManager(
            status_progress=self.status_progress,
            status_message=self.status_message,
            threadpool=self.threadpool,
            interrupt_on=self.settings_processing_widget.interrupt_signal,
        )
        self.nn_manager.picking_finished_signal.connect(self.on_picking_finished)
        self.nn_manager.picking_not_started_error_signal.connect(self.on_picking_not_started_error)

        self.is_toggled_picks_from_file = False
        # placeholders
        self.sgy: Optional[SGY] = None
        self.fn_sgy: Optional[Union[str, Path]] = None
        self.ready_to_process = ReadyToProcess()
        self.last_task: Optional[Task] = None
        self.settings: Optional[Dict[str, Any]] = None
        self.last_folder: Optional[Union[str, Path]] = None
        self.picks_from_file_in_ms: Optional[Tuple[Union[int, float], ...]] = None
        self.picker_hash = MODEL_ONNX_HASH

        self.show()

    def get_last_folder(self) -> str:
        if self.last_folder is None:
            return str(Path.home())
        else:
            if Path(self.last_folder).exists():
                return str(Path(self.last_folder).resolve())
            else:
                return str(Path.home())

    def set_last_folder_based_on_file(self, file: Union[str, Path]) -> None:
        file = Path(file)
        if file.exists():
            self.last_folder = str(file.parent)
        else:
            self.last_folder = None

    def pick_fb(self, settings: PickingSettings) -> None:
        self.button_get_filename.setEnabled(False)
        self.nn_manager.pick_fb(self.sgy, settings)

    def on_picking_not_started_error(self, exc: ExceptionOptional) -> None:
        self.settings_processing_widget.set_selection_mode()
        window_error = MessageBox(
            self,
            title=exc.exception.__class__.__name__,
            message=str(exc),
            detailed_message=exc.get_formatted_traceback(),
        )
        window_error.exec_()

    def on_picking_finished(self, result: Task) -> None:
        self.settings_processing_widget.set_selection_mode()
        self.last_task = result

        if result.success:
            self.graph.plot_nn_picks(self.last_task.picks_in_ms)
            self.run_processing_region()
            self.button_export.setEnabled(True)
        else:
            if isinstance(result.exception, InterruptedError):
                window_error = MessageBox(
                    self,
                    title="Interruption",
                    message="The picking process has been interrupted. Intermediate results will not be saved",
                )
            else:
                window_error = MessageBox(
                    self,
                    title=result.exception.__class__.__name__,
                    message=result.error_message,
                    detailed_message=result.get_formatted_traceback(),
                )
            window_error.exec_()

        self.button_get_filename.setEnabled(True)

    def processing_region_changed(self, toggle: bool) -> None:
        self.need_processing_region = toggle
        self.run_processing_region()

    def run_processing_region(self) -> None:
        if self.need_processing_region:
            self.show_processing_region()
        else:
            self.hide_processing_region()

    def show_processing_region(self) -> None:
        if self.last_task and self.last_task.success:
            self.graph.plot_processing_region(self.last_task.traces_per_gather, self.last_task.maximum_time)

    def hide_processing_region(self) -> None:
        if self.last_task and self.last_task.success:
            self.graph.remove_processing_region()

    def show_nn_picks(self) -> None:
        if self.last_task and self.last_task.success:
            self.graph.plot_nn_picks(self.last_task.picks_in_ms)

    def read_picks_from_file(self) -> None:
        picks = self.sgy.read_custom_trace_header(
            **TraceHeaderParams(**self.picks_from_file_settings.model_dump()).model_dump(),
        )
        units_converter = UnitsConverter(sgy_mcs=self.sgy.dt_mcs)
        if self.picks_from_file_settings.picks_unit == "sample":
            self.picks_from_file_in_ms = units_converter.index2ms(picks)  # type: ignore
        elif self.picks_from_file_settings.picks_unit == "mcs":
            self.picks_from_file_in_ms = units_converter.mcs2ms(picks)  # type: ignore
        else:
            self.picks_from_file_in_ms = picks

    def update_picks_from_file_settings(self, new_settings: PicksFromFileSettings) -> None:
        print(new_settings)
        self.picks_from_file_settings = new_settings

    def toggle_picks_from_file(self, toggled: bool) -> None:
        self.is_toggled_picks_from_file = toggled
        self.show_picks_from_file()

    def show_picks_from_file(self) -> None:
        if self.is_toggled_picks_from_file:
            self.read_picks_from_file()
            self.graph.plot_extra_picks(picks_ms=self.picks_from_file_in_ms, color=(0, 0, 255))
        else:
            self.graph.remove_extra_picks()

    def update_plotseis_settings(self, new_settings: PlotseisSettings) -> None:
        self.plotseis_settings = new_settings
        self.update_plot(False)

    def update_plot(self, refresh_view: bool = False) -> None:
        self.graph.plotseis(self.sgy, refresh_view=refresh_view, **self.plotseis_settings.model_dump())
        self.show_processing_region()
        self.show_nn_picks()
        self.show_picks_from_file()

    def show_settings_processing_window(self) -> None:
        self.settings_processing_widget.show()
        self.settings_processing_widget.focusWidget()

    def unlock_pickng_if_ready(self) -> None:
        if self.ready_to_process.is_ready():
            self.button_settings_processing.setEnabled(True)
            self.settings_processing_widget.enable_only_visualizations_settings()
            self.settings_processing_widget.enable_picking()
            self.status_message.setText("Click on picking to start processing")

    def load_nn(self, filename: Optional[Union[str, Path]] = None) -> None:
        if not filename:
            options = QFileDialog.Options()
            filename, _ = QFileDialog.getOpenFileName(
                self, "Select file with NN weights", directory=self.get_last_folder(), options=options
            )

        if filename:
            if FileState.get_file_state(filename, self.picker_hash) == FileState.valid_file:
                self.nn_manager.init_net(weights=filename)
                self.button_load_nn.setEnabled(False)
                self.ready_to_process.model_loaded = True

                status_message = "Model loaded successfully"
                if not self.ready_to_process.sgy_selected:
                    status_message += ". Open SGY file to start picking"
                self.status_message.setText(status_message)

                self.unlock_pickng_if_ready()
                self.set_last_folder_based_on_file(filename)
            else:
                window_err = MessageBox(
                    self,
                    title="Model loading error",
                    message="The file cannot be used as model weights. "
                    "Download the file according to the manual and select it.",
                )
                window_err.exec_()

    def get_filename(self, filename: Optional[Union[str, Path]] = None) -> None:
        if not filename:
            filename, _ = QFileDialog.getOpenFileName(
                self,
                "Open SEGY-file",
                directory=self.get_last_folder(),
                filter="SEGY-file (*.segy *.sgy);; Any file (*)",
            )
        if filename:
            try:
                self.fn_sgy = Path(filename)
                self.last_task = None
                self.sgy = SGY(self.fn_sgy)
                self.picks_from_file_in_ms = None

                self.graph.full_clean()
                self.update_plot(refresh_view=True)
                self.graph.show()
                self.button_export.setEnabled(False)
                self.button_settings_processing.setEnabled(True)
                self.settings_processing_widget.enable_only_visualizations_settings()

                self.button_get_filename.setEnabled(True)
                self.ready_to_process.sgy_selected = True

                if not self.ready_to_process.model_loaded:
                    status_message = "Load model to start picking"
                    self.status_message.setText(status_message)

                self.unlock_pickng_if_ready()
                self.set_last_folder_based_on_file(filename)

            except Exception as e:
                window_err = MessageBox(self, title=e.__class__.__name__, message=str(e))
                window_err.exec_()

    def export(self) -> None:
        formats = ["SEGY-file (*.segy *.sgy)", "JSON-file (*.json)", "TXT-file (*.txt)"]
        formats = ";; ".join(formats)
        filename, _ = QFileDialog.getSaveFileName(self, "Save result", directory=self.get_last_folder(), filter=formats)

        if filename:
            filename = Path(filename).resolve()
            if self.last_task is not None and self.last_task.success:
                filename.parent.mkdir(parents=True, exist_ok=True)

                picks_in_samples_prev = self.last_task.picks_in_samples
                if self.graph.is_picks_modified_manually:
                    self.last_task.picks_in_samples = multiply_iterable_by(
                        self.graph.nn_picks_in_ms, 1 / self.sgy.dt_ms, int
                    )
                if filename.suffix.lower() in (".sgy", ".segy"):
                    save_params = QDialogByteEncodeUnit(first_byte=1, byte_position=237, encoding="I", picks_unit="mcs")
                    save_params.setWindowTitle("How to save first breaks")
                    save_params.show()
                    save_params.exec_()
                    export_params = save_params.get_values()
                    self.last_task.export_result_as_sgy(str(filename), **export_params)  # type: ignore
                elif filename.suffix.lower() == ".txt":
                    self.last_task.export_result_as_txt(str(filename))
                elif filename.suffix.lower() == ".json":
                    self.last_task.export_result_as_json(str(filename))
                else:
                    message_er = "The file can only be saved in '.sgy', '.segy', '.txt, or '.json' formats"
                    window_err = MessageBox(self, title="Wrong filename", message=message_er)
                    window_err.exec_()
                if self.graph.is_picks_modified_manually:
                    self.last_task.picks_in_samples = picks_in_samples_prev

    def closeEvent(self, e: QCloseEvent) -> None:
        self.graph.spectrum_window.close()
        e.accept()


def run_app() -> None:
    app = QApplication([])
    _ = MainWindow()
    app.exec_()


def fetch_data_and_run_app() -> None:
    download_model_onnx(MODEL_ONNX_PATH)
    download_demo_sgy(DEMO_SGY_PATH)
    app = QApplication([])
    window = MainWindow()
    window.load_nn(MODEL_ONNX_PATH)
    window.get_filename(DEMO_SGY_PATH)
    app.exec_()


if __name__ == "__main__":
    fetch_data_and_run_app()



from typing import Any, Dict, List, Optional, Tuple, Union

from PyQt5 import QtWidgets
from PyQt5.QtCore import QEvent, QPoint, QPointF, Qt
from PyQt5.QtWidgets import (
    QDesktopWidget,
    QDialog,
    QDialogButtonBox,
    QGraphicsSceneMouseEvent,
    QLabel,
    QPushButton,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)
from pyqtgraph import ViewBox


class MessageBox(QDialog):
    def __init__(
        self,
        parent: QWidget,
        title: str = "Error",
        message: str = "Error",
        detailed_message: str = None,
        add_cancel_option: bool = False,
        label_ok: str = "Ok",
        label_cancel: str = "Cancel",
    ):
        super().__init__(parent=parent)

        self.setWindowTitle(title)
        self.setWindowModality(Qt.ApplicationModal)
        set_geometry(self, width_rel=0.2, height_rel=0.2, centralize=True, fix_size=False)

        error_label = QLabel(message)
        error_label.setWordWrap(True)
        error_label.setTextInteractionFlags(Qt.LinksAccessibleByMouse | Qt.TextSelectableByMouse)

        self.button_box = QDialogButtonBox()
        self.button_box.addButton(label_ok, QDialogButtonBox.AcceptRole)
        self.button_box.accepted.connect(self.accept)
        if add_cancel_option:
            self.button_box.addButton(label_cancel, QDialogButtonBox.RejectRole)
            self.button_box.rejected.connect(self.reject)

        self.layout = QVBoxLayout()
        self.layout.addWidget(error_label)

        if detailed_message:
            self.details_button = QPushButton("Show Details")
            self.details_text = QTextEdit()
            self.details_text.setPlainText(detailed_message)
            self.details_text.setVisible(False)
            self.details_text.setReadOnly(True)

            self.details_button.clicked.connect(self.toggle_details)

            self.layout.addWidget(self.details_button)
            self.layout.addWidget(self.details_text)

        self.layout.addWidget(self.button_box)
        self.setLayout(self.layout)

    def toggle_details(self) -> None:
        if self.details_text.isVisible():
            self.details_text.setVisible(False)
            self.details_button.setText("Show Details")
        else:
            self.details_text.setVisible(True)
            self.details_button.setText("Hide Details")


def set_geometry(
    widget: QWidget,
    width_rel: float,
    height_rel: float,
    centralize: bool = True,
    fix_size: bool = False,
) -> None:
    assert 0 < width_rel <= 1
    assert 0 < height_rel <= 1
    h, w = widget.screen().size().height(), widget.screen().size().width()
    width = int(width_rel * w)
    height = int(height_rel * h)
    widget.setGeometry(0, 0, width, height)

    if centralize:
        qt_rectangle = widget.frameGeometry()
        center_point = QDesktopWidget().availableGeometry().center()
        qt_rectangle.moveCenter(center_point)
        widget.move(qt_rectangle.topLeft())

    if fix_size:
        widget.setFixedSize(width, height)

    monitor = QDesktopWidget().screenGeometry(1)
    widget.move(monitor.left(), monitor.top())


class QHSeparationLine(QtWidgets.QWidget):
    def __init__(self, text: str = ""):
        super().__init__()

        # Create the horizontal line (QFrame)
        self.line = QtWidgets.QFrame(self)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)

        # Create the label for the text
        self.label = QtWidgets.QLabel(text, self)
        self.label.setStyleSheet("background-color: transparent; color: grey;")
        self.label.setAlignment(Qt.AlignCenter)

        # Create a horizontal layout to hold the line and the label
        layout = QtWidgets.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.line, 1)
        layout.addWidget(self.label, 0)
        layout.addWidget(self.line, 1)

        # Adjust margins and spacing
        layout.setSpacing(10)  # space between label and line
        layout.setContentsMargins(0, 10, 0, 0)  # top margin to position label above line

        self.setLayout(layout)


# class QHSeparationLine(QtWidgets.QFrame):
#     def __init__(self) -> None:
#         super().__init__()
#         self.setMinimumWidth(1)
#         self.setFixedHeight(20)
#         self.setFrameShape(QtWidgets.QFrame.HLine)
#         self.setFrameShadow(QtWidgets.QFrame.Sunken)
#         self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)


TMappingSetup = Dict[int, Union[Tuple[str, Any], List[Any]]]


def validate_mapping_setup_and_get_current_index(
    mapping: TMappingSetup,
    current_index: Optional[int] = None,
    current_label: Optional[str] = None,
    current_value: Optional[Any] = None,
) -> int:
    assert all(isinstance(k, int) for k in mapping.keys())
    assert all(isinstance(v, (list, tuple)) and len(v) == 2 for v in mapping.values())
    assert all(isinstance(label, str) for label, _ in mapping.values())
    assert set(range(len(mapping.keys()))) == set(mapping.keys())

    current_is_not_none = [v is not None for v in [current_index, current_label, current_value]]
    assert (
        0 <= sum(current_is_not_none) <= 1
    ), "Only one of 'current_index' or 'current_label' or 'current_value' might be specified."

    if current_label or current_value:
        if current_label:
            assert isinstance(current_label, str)
            idx_to_check = 0
            type_to_check = "Label"
            got = current_label
        else:
            idx_to_check = 1
            type_to_check = "Value"
            got = current_value

        corresponding_ids = [k for k, v in mapping.items() if v[idx_to_check] == got]

        if not corresponding_ids:
            raise KeyError(
                f"{type_to_check} '{type_to_check}' not found in mapping. "
                f"Available {[v[idx_to_check] for v in mapping.values()]}, got '{got}'"
            )
        current_index = corresponding_ids[0]
    elif current_index:
        assert isinstance(current_index, int)
        assert 0 <= current_index < len(mapping)
    else:
        current_index = 0

    return current_index


def get_mouse_position_in_scene_coords(
    event_or_point: Union[QGraphicsSceneMouseEvent, QEvent, QPointF, QPoint], viewbox: ViewBox
) -> QPointF:
    if isinstance(event_or_point, QGraphicsSceneMouseEvent):
        point = event_or_point.scenePos()
    elif isinstance(event_or_point, QEvent):
        point = event_or_point.localPos()
    elif isinstance(event_or_point, (QPointF, QPoint)):
        point = event_or_point
    else:
        raise TypeError("Only events and points are available")
    return viewbox.mapSceneToView(point)



from typing import Any, Optional

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QApplication, QHBoxLayout, QLabel, QSlider, QWidget


class QSliderWithValues(QWidget):
    value_changed_signal = pyqtSignal(float)
    slider_released_signal = pyqtSignal()
    value_on_released_signal = pyqtSignal(float)

    def __init__(
        self,
        value: float = 1,
        min_value: float = -1,
        max_value: float = 1,
        step: float = 0.1,
        decimals: int = 1,
        ticks_interval: Optional[float] = None,
        margins: Optional[int] = None,
        slider_space_fraction: Optional[float] = None,
        block_mouse_scrolling: bool = True,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)

        self.layout = QHBoxLayout()
        if margins is not None:
            self.layout.setContentsMargins(margins, margins, margins, margins)
        self.setLayout(self.layout)

        self.slider_value = value
        self.min_value = min_value
        self.max_value = max_value
        self.step = step
        self.decimals = decimals
        self.num_steps = int((self.max_value - self.min_value) / self.step)

        self.slider_label = QLabel(str(self.slider_value), self)

        self.slider_widget = QSlider(Qt.Horizontal, self)

        self.slider_widget.setRange(0, self.num_steps)
        self.slider_widget.setValue(self.value2slider(self.slider_value))

        if slider_space_fraction is not None:
            slider_space = int(slider_space_fraction * 100)
            self.layout.addWidget(self.slider_widget, slider_space)
            self.layout.addWidget(self.slider_label, 100 - slider_space)
        else:
            self.layout.addWidget(self.slider_widget)
            self.layout.addWidget(self.slider_label)

        if ticks_interval is not None:
            self.slider_widget.setTickInterval(int(ticks_interval / self.step))
            self.slider_widget.setTickPosition(QSlider.TicksAbove)
            self.slider_widget.setSingleStep(1)

        if block_mouse_scrolling:
            self.slider_widget.wheelEvent = lambda *e_args: e_args[-1].ignore()  # block scrolling with wheel

        self.slider_widget.sliderPressed.connect(self.slider_pressed)
        self.slider_widget.sliderReleased.connect(self.slider_released)
        self.slider_widget.sliderMoved.connect(self.value_changed)

        self.show()

    def slider_pressed(self) -> None:
        self.value_changed()

    def slider_released(self) -> None:
        self.value_on_released_signal.emit(self.slider_value)
        self.slider_released_signal.emit()

    def value_changed(self, *args: Any) -> None:
        self.slider_value = self.slider2value(self.slider_widget.value())
        self.slider_label.setText(self.text())
        self.value_changed_signal.emit(self.slider_value)

    def text(self) -> str:
        return f"{self.slider_value:.{self.decimals}f}"

    def value2slider(self, value_float: float) -> int:
        return int((value_float - self.min_value) / self.step)

    def slider2value(self, value_from_slider: int) -> float:
        return self.min_value + (value_from_slider * self.step)

    def value(self) -> float:
        return self.slider_value


if __name__ == "__main__":
    app = QApplication([])
    window = QSliderWithValues(1, -10, 10, slider_space_fraction=0.7, margins=20, ticks_interval=5)
    # window = QSlider(Qt.Horizontal)
    # window.valueChanged.connect(lambda x: print(x))
    # window.rele
    window.show()
    app.exec_()



from typing import Any, Optional

from tqdm.auto import tqdm

from first_breaks.picking.task import Task


class IPicker:
    def __init__(self, show_progressbar: bool = True, *args: Any, **kwargs: Any):
        self.show_progressbar = show_progressbar
        self.progressbar: Optional[tqdm] = None
        self.need_interrupt = False

    def change_settings(self, *args: Any, **kwargs: Any) -> None:
        pass

    def process_task(self, task: Task) -> Task:
        raise NotImplementedError

    def callback_interrupt(self) -> None:
        self.need_interrupt = True

    def interrupt_if_need(self) -> None:
        if self.need_interrupt:
            self.need_interrupt = False
            raise InterruptedError("Processing interrupted")

    def callback_processing_started(self, length: int) -> Any:
        if self.show_progressbar:
            self.progressbar = tqdm(desc="Picking", total=length)

    def callback_processing_finished(self) -> Any:
        if self.show_progressbar:
            self.progressbar.close()

    def callback_step_finished(self, finished_step: int) -> Any:
        if self.show_progressbar:
            interval = finished_step - self.progressbar.n
            self.progressbar.update(interval)



import colorsys
import hashlib
import inspect
import io
import random
from itertools import islice
from pathlib import Path
from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple, Union

import numpy as np
import requests

from first_breaks.const import (
    DEMO_SGY_HASH,
    DEMO_SGY_PATH,
    DEMO_SGY_URL,
    MODEL_ONNX_HASH,
    MODEL_ONNX_PATH,
    MODEL_ONNX_URL,
    TIMEOUT,
)

TScalar = Union[int, float, np.number]
TTimeType = Union[TScalar, List[TScalar], Tuple[TScalar, ...], np.ndarray]


class InvalidHash(Exception):
    pass


def chunk_iterable(it: Iterable[Any], size: int) -> List[Tuple[Any, ...]]:
    it = iter(it)
    return list(iter(lambda: tuple(islice(it, size)), ()))


def get_io(source: Union[Path, str, bytes], mode: str = "r") -> Union[io.BytesIO, io.FileIO]:
    if isinstance(source, (Path, str)):
        source = Path(source).resolve()
        if "r" in mode:
            if not source.exists():
                raise FileNotFoundError("There is no file: ", str(source))
        descriptor = io.FileIO(str(source), mode=mode)
    elif isinstance(source, bytes):
        descriptor = io.BytesIO(source)  # type: ignore
    else:
        raise TypeError("Not supported type")
    return descriptor


def calc_hash(source: Union[Path, str, bytes, io.BytesIO, io.FileIO]) -> str:
    hash_md5 = hashlib.md5()
    if not isinstance(source, (io.BytesIO, io.FileIO)):
        source = get_io(source, mode="rb")
    for chunk in iter(lambda: source.read(4096), b""):  # type: ignore
        hash_md5.update(chunk)
    return hash_md5.hexdigest()


def download_by_url(url: str, fname: Optional[Union[str, Path]], timeout: float = TIMEOUT) -> Optional[bytes]:
    response = requests.get(url, timeout=timeout)
    if response.status_code != 200:
        response.raise_for_status()
        return None
    else:
        if fname:
            Path(fname).parent.mkdir(exist_ok=True, parents=True)
            with open(fname, "wb+") as f:
                f.write(response.content)
        return response.content


def download_and_validate_file(
    fname: Union[str, Path], url: str, md5: str, timeout: float = TIMEOUT
) -> Union[str, Path]:
    if not (Path(fname).exists() and calc_hash(fname) == md5):
        download_by_url(url=url, fname=fname, timeout=timeout)
    md5_last = calc_hash(fname)
    if md5_last != md5:
        raise InvalidHash(f"Hash for file {Path(fname).resolve()} in invalid. Got {md5_last}, expected {md5}")
    return fname


def download_demo_sgy(
    fname: Union[str, Path] = DEMO_SGY_PATH, url: str = DEMO_SGY_URL, md5: str = DEMO_SGY_HASH
) -> Union[str, Path]:
    return download_and_validate_file(fname=fname, url=url, md5=md5)


def download_model_onnx(
    fname: Union[str, Path] = MODEL_ONNX_PATH, url: str = MODEL_ONNX_URL, md5: str = MODEL_ONNX_HASH
) -> Union[str, Path]:
    return download_and_validate_file(fname=fname, url=url, md5=md5)


def multiply_iterable_by(sample: TTimeType, multiplier: float, cast_to: Optional[Any] = None) -> TTimeType:
    if isinstance(sample, (int, float, str)):
        result = sample * multiplier  # type: ignore
        return cast_to(result) if cast_to is not None else result
    elif isinstance(sample, (np.number, np.ndarray)):
        result = sample * multiplier
        return result.astype(cast_to) if cast_to is not None else result
    elif isinstance(sample, list):
        return list(multiply_iterable_by(val, multiplier, cast_to) for val in sample)
    elif isinstance(sample, tuple):
        return tuple(multiply_iterable_by(val, multiplier, cast_to) for val in sample)
    else:
        raise TypeError("Invalid type for samples")


class UnitsConverter:
    def __init__(
        self,
        *args: Any,
        sgy_mcs: Optional[Union[int, float]] = None,
        sgy_ms: Optional[Union[int, float]] = None,
    ):
        if args:
            raise ValueError("Specify explicitly either `sgy_mcs`or `sgy_ms` as keyword argument")
        if (sgy_mcs is None and sgy_ms is None) or (sgy_mcs is not None and sgy_ms is not None):
            raise RuntimeError("One and only one of `sgy_mcs` or `sgy_ms` must be specified")
        elif sgy_mcs is not None:
            self.sgy_mcs = sgy_mcs
            self.sgy_ms = self.mcs2ms(sgy_mcs)  # type: ignore
        elif sgy_ms is not None:
            self.sgy_mcs = self.ms2mcs(sgy_ms)  # type: ignore
            self.sgy_ms = sgy_ms
        else:
            raise RuntimeError("Init error")

    @staticmethod
    def ms2mcs(sample: TTimeType, cast_to: Any = int) -> TTimeType:
        return multiply_iterable_by(sample, 1000, cast_to)

    @staticmethod
    def mcs2ms(sample: TTimeType, cast_to: Any = float) -> TTimeType:
        return multiply_iterable_by(sample, 0.001, cast_to)

    def ms2index(self, sample: TTimeType, cast_to: Any = int) -> TTimeType:
        return multiply_iterable_by(sample, 1 / self.sgy_ms, cast_to)  # type: ignore

    def mcs2index(self, sample: TTimeType, cast_to: Any = int) -> TTimeType:
        return multiply_iterable_by(sample, 1 / self.sgy_mcs, cast_to)

    def index2ms(self, sample: TTimeType, cast_to: Any = float) -> TTimeType:
        return multiply_iterable_by(sample, self.sgy_ms, cast_to)  # type: ignore

    def index2mcs(self, sample: TTimeType, cast_to: Any = int) -> TTimeType:
        return multiply_iterable_by(sample, self.sgy_mcs, cast_to)


def remove_unused_kwargs(kwargs: Dict[str, Any], constructor: Any) -> Dict[str, Any]:
    return {k: v for k, v in kwargs.items() if k in inspect.signature(constructor).parameters}


def _color_generator() -> Generator[List[int], None, None]:
    golden_ratio = 0.618033988749895
    hue = random.random()  # start from a random position
    while True:
        hue += golden_ratio
        hue %= 1
        yield [int(255 * v) for v in colorsys.hsv_to_rgb(hue, 0.5, 0.95)]


cgen = _color_generator()


def generate_color() -> List[int]:
    return next(cgen)


def resolve_postime2xy(vsp_view: bool, position: Any, time: Any) -> Tuple[Any, Any]:
    if vsp_view:
        return time, position
    else:
        return position, time


def resolve_xy2postime(vsp_view: bool, x: Any, y: Any) -> Tuple[Any, Any]:
    if vsp_view:
        return y, x
    else:
        return x, y



import warnings
from typing import Optional

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtGui import QDoubleValidator
from PyQt5.QtWidgets import (
    QApplication,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QLabel,
    QLineEdit,
    QSpinBox,
    QWidget,
)

from first_breaks.const import HIGH_DPI
from first_breaks.desktop.utils import MessageBox, QHSeparationLine, set_geometry
from first_breaks.picking.task import Task
from first_breaks.utils.cuda import ONNX_CUDA_AVAILABLE, get_recommended_device

if HIGH_DPI:
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)

warnings.filterwarnings("ignore")


class PickingWindow(QDialog):

    export_settings_signal = pyqtSignal(dict)

    def __init__(
        self,
        task: Optional[Task] = None,
        device: str = get_recommended_device(),
        batch_size: int = 1,
    ):
        assert device in ["cuda", "cpu"]
        super().__init__()

        self.setWindowTitle("Picking settings")
        self.setWindowModality(Qt.ApplicationModal)
        set_geometry(self, width_rel=0.22, height_rel=0.3, fix_size=True, centralize=True)

        self.storage = {}

        self.layout = QGridLayout()
        self.setLayout(self.layout)

        self.traces_per_gather_label = QLabel("Traces per gather")
        self.traces_per_gather = QSpinBox()
        self.storage["traces_per_gather"] = [
            self.traces_per_gather_label,
            self.traces_per_gather,
            self.get_current_value_default,
            int,
        ]
        self.traces_per_gather.setRange(2, 999)
        value_traces_per_gather = task.traces_per_gather if task else 48
        self.traces_per_gather.setValue(value_traces_per_gather)
        self.layout.addWidget(self.traces_per_gather_label, 0, 0)
        self.layout.addWidget(self.traces_per_gather, 0, 1)

        self.maximum_time_label = QLabel("Maximum time, ms")
        self.maximum_time = QLineEdit()
        self.storage["maximum_time"] = [
            self.maximum_time_label,
            self.maximum_time,
            self.get_current_value_default,
            float,
        ]
        maximum_time_validator = QDoubleValidator()
        maximum_time_validator.setBottom(0.0)
        self.maximum_time.setValidator(maximum_time_validator)
        value_maximum_time = task.maximum_time if task else 100.0
        self.maximum_time.setText(str(value_maximum_time))
        self.layout.addWidget(self.maximum_time_label, 1, 0)
        self.layout.addWidget(self.maximum_time, 1, 1)

        self.gain_label = QLabel("Gain")
        self.gain = QLineEdit()
        self.storage["gain"] = [self.gain_label, self.gain, self.get_current_value_default, float]
        gain_validator = QDoubleValidator()
        self.gain.setValidator(gain_validator)
        value_gain = task.gain if task else 1
        self.gain.setText(str(value_gain))
        self.layout.addWidget(self.gain_label, 2, 0)
        self.layout.addWidget(self.gain, 2, 1)

        self.clip_label = QLabel("Clip")
        self.clip = QLineEdit()
        self.storage["clip"] = [self.clip_label, self.clip, self.get_current_value_default, float]
        clip_validator = QDoubleValidator()
        clip_validator.setBottom(0.1)
        self.clip.setValidator(clip_validator)
        value_clip = task.clip if task else 1
        self.clip.setText(str(value_clip))
        self.layout.addWidget(self.clip_label, 3, 0)
        self.layout.addWidget(self.clip, 3, 1)

        self.layout.addWidget(QHSeparationLine(), 4, 0, 1, 2)

        self.device_label = QLabel("Device")
        self.device = QComboBox()
        self.storage["device"] = [self.device_label, self.device, self.get_current_value_combobox, str]
        self.device_idx2labelvalue = {0: ["GPU/CUDA", "cuda"], 1: ["CPU", "cpu"]}
        self.device.addItem("GPU/CUDA")
        self.device.addItem("CPU")
        if device == "cuda" and ONNX_CUDA_AVAILABLE:
            self.device.setCurrentIndex(0)
        else:
            self.device.setCurrentIndex(1)
        if not ONNX_CUDA_AVAILABLE:
            self.device.setEnabled(False)
        self.device.activated.connect(self.set_enabled_batch_size_depend_on_device)
        self.layout.addWidget(self.device_label, 5, 0)
        self.layout.addWidget(self.device, 5, 1)

        self.batch_size_label = QLabel("Batch size (GPU)")
        self.batch_size = QSpinBox()
        self.storage["batch_size"] = [self.batch_size_label, self.batch_size, self.get_current_value_default, int]
        self.batch_size.setRange(1, 20)
        self.batch_size.setValue(batch_size)
        self.set_enabled_batch_size_depend_on_device()
        self.layout.addWidget(self.batch_size_label, 6, 0)
        self.layout.addWidget(self.batch_size, 6, 1)

        self.button_box = QDialogButtonBox()
        self.button_box.addButton("Run picking", QDialogButtonBox.AcceptRole)
        self.button_box.addButton("Cancel", QDialogButtonBox.RejectRole)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        self.layout.addWidget(self.button_box)
        self.show()

    def set_enabled_batch_size_depend_on_device(self, _: Optional[int] = None) -> None:
        current_device = self.get_current_value_combobox(self.device)
        if current_device == "cpu":
            self.batch_size.setEnabled(False)
        else:
            self.batch_size.setEnabled(True)

    @staticmethod
    def get_current_value_default(widget: QWidget) -> str:
        return widget.text()

    def get_current_value_combobox(self, widget: QComboBox) -> str:
        return self.device_idx2labelvalue[widget.currentIndex()][1]

    def accept(self) -> None:
        empty_fields = [
            label.split(",")[0] for label, widget, getter, _ in self.storage.values() if not str(getter(widget)).strip()
        ]

        if empty_fields:
            template = "Fields" if len(empty_fields) > 1 else "Field"
            template += " {} must be filled"
            invalid_fields_str = ",".join(empty_fields)
            window_error = MessageBox(self, title="Input error", message=template.format(invalid_fields_str))
            window_error.exec_()
        else:
            settings = {name: cast(getter(widget)) for name, (_, widget, getter, cast) in self.storage.items()}
            print(settings)
            self.export_settings_signal.emit(settings)
            super().accept()

    def reject(self) -> None:
        self.export_settings_signal.emit({})
        super().reject()


if __name__ == "__main__":
    app = QApplication([])
    window = PickingWindow()
    app.exec_()



import warnings
from typing import Any, Dict, Optional, Tuple

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import (
    QApplication,
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from first_breaks.const import HIGH_DPI
from first_breaks.desktop.combobox_with_mapping import QComboBoxMapping
from first_breaks.sgy.headers import Headers

if HIGH_DPI:
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)

warnings.filterwarnings("ignore")


def build_encoding_mapping() -> Dict[int, Tuple[str, str]]:
    mapping = {
        0: ("Integer 4", "i"),
        1: ("UInteger 4", "I"),
        2: ("Short 2", "h"),
        3: ("UShort 2", "H"),
        4: ("Long 4", "l"),
        5: ("ULong 4", "L"),
        6: ("Float 4", "f"),
        7: ("Double 8", "d"),
    }
    assert all(v[1] in Headers.format2size.keys() for v in mapping.values())
    return mapping


ENCODING_MAPPING = build_encoding_mapping()


class QByteEncodeUnitWidget(QWidget):
    values_changed_signal = pyqtSignal(dict)

    def __init__(
        self,
        byte_position: int = 0,
        encoding: str = "I",
        first_byte: int = 0,
        picks_unit: str = "mcs",
        margins: Optional[int] = None,
        *args: Any,
        **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)

        self.layout = QHBoxLayout()
        if margins is not None:
            self.layout.setContentsMargins(margins, margins, margins, margins)
        self.setLayout(self.layout)

        assert byte_position >= first_byte
        self.byte_position_value = byte_position
        self.encoding_value = encoding
        self.first_byte = first_byte
        self.byte_position_maximum = 240 + self.first_byte
        self.picks_unit_value = picks_unit

        self.byte_position_label = QLabel("Byte")
        self.byte_position_widget = QSpinBox()
        self.byte_position_widget.setRange(self.first_byte, self.byte_position_maximum)
        self.byte_position_widget.setValue(self.byte_position_value)
        self.byte_position_widget.valueChanged.connect(self.values_changed)

        self.layout.addWidget(self.byte_position_label)
        self.layout.addWidget(self.byte_position_widget)

        self.encoding_label = QLabel("Encoding")
        self.encoding_widget = QComboBoxMapping(ENCODING_MAPPING, current_value=self.encoding_value)  # type: ignore

        self.layout.addWidget(self.encoding_label)
        self.layout.addWidget(self.encoding_widget)

        self.picks_unit_label = QLabel("Unit")
        self.picks_unit_widget = QComboBoxMapping(
            {0: ("Microseconds", "mcs"), 1: ("Milliseconds", "ms"), 2: ("Samples", "sample")},
            current_value=self.picks_unit_value,
        )

        self.layout.addWidget(self.picks_unit_label)
        self.layout.addWidget(self.picks_unit_widget)

        self._previous_values = self.get_values()
        self.align_and_update_values()

        # connect after aligning
        self.picks_unit_widget.value_changed_signal.connect(self.values_changed)
        self.encoding_widget.value_changed_signal.connect(self.values_changed)

    def align_and_update_values(self) -> None:
        self.encoding_value = self.encoding_widget.value()
        self.byte_position_value = self.byte_position_widget.value()
        self.picks_unit_value = self.picks_unit_widget.value()

        byte_position_maximum_aligned = self.first_byte + 240 - Headers.format2size[self.encoding_value]
        byte_position_value_aligned = min(self.byte_position_value, byte_position_maximum_aligned)

        if byte_position_maximum_aligned != self.byte_position_maximum:
            self.byte_position_maximum = byte_position_maximum_aligned
            self.byte_position_widget.setMaximum(self.byte_position_maximum)

        if byte_position_value_aligned != self.byte_position_value:
            self.byte_position_value = byte_position_value_aligned
            self.byte_position_widget.setValue(self.byte_position_value)

    def get_values(self) -> Dict[str, Any]:
        return {
            "byte_position": self.byte_position_widget.value() - self.first_byte,
            "encoding": self.encoding_widget.value(),
            "picks_unit": self.picks_unit_widget.value(),
        }

    def values_changed(self) -> None:
        self.align_and_update_values()
        values = self.get_values()
        if values != self._previous_values:  # avoid emit extra signal if chained triggers happen
            self._previous_values = values
            self.values_changed_signal.emit(values)


class QDialogByteEncodeUnit(QDialog):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__()

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        self.widget = QByteEncodeUnitWidget(*args, **kwargs)
        self.layout.addWidget(self.widget)

        self.button_box = QDialogButtonBox()
        self.button_box.addButton("Ok", QDialogButtonBox.AcceptRole)
        self.button_box.accepted.connect(self.accept)

        self.layout.addWidget(self.button_box)

    def get_values(self) -> Dict[str, Any]:
        return self.widget.get_values()


if __name__ == "__main__":
    app = QApplication([])
    window = QDialogByteEncodeUnit(first_byte=1, byte_position=237, encoding="I", picks_unit="mcs")
    window.show()
    app.exec_()
    print(window.get_values())



import warnings
from typing import Any, Optional

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QApplication, QComboBox

from first_breaks.const import HIGH_DPI
from first_breaks.desktop.utils import (
    TMappingSetup,
    validate_mapping_setup_and_get_current_index,
)

if HIGH_DPI:
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)

warnings.filterwarnings("ignore")


class QComboBoxMapping(QComboBox):
    index_changed_signal = pyqtSignal(int)
    value_changed_signal = pyqtSignal(object)
    label_changed_signal = pyqtSignal(str)
    changed_signal = pyqtSignal()

    def __init__(
        self,
        mapping: TMappingSetup,
        current_index: Optional[int] = None,
        current_label: Optional[str] = None,
        current_value: Optional[str] = None,
        *args: Any,
        **kwargs: Any
    ):
        super().__init__(*args, **kwargs)

        current_index = validate_mapping_setup_and_get_current_index(
            mapping,
            current_index=current_index,
            current_label=current_label,
            current_value=current_value,
        )

        self.mapping = mapping
        self.current_index = current_index

        items = [v[0] for v in self.mapping.values()]
        self.addItems(items)

        self.setCurrentIndex(current_index)
        self.currentIndexChanged.connect(self.changed)

    def changed(self) -> None:
        self.index_changed_signal.emit(self.currentIndex())
        self.value_changed_signal.emit(self.value())
        self.label_changed_signal.emit(self.label())
        self.changed_signal.emit()

    def value(self) -> str:
        return self.mapping[self.currentIndex()][1]

    def text(self) -> str:
        return str(self.value())

    def label(self) -> str:
        return self.mapping[self.currentIndex()][0]



from pathlib import Path
from typing import Any, Literal, Optional, Union

from pydantic import Field, field_validator, model_validator
from pydantic_core.core_schema import FieldValidationInfo

from first_breaks.data_models.independent import DefaultModel, TraceBytePosition
from first_breaks.sgy.headers import Headers, TraceHeaders
from first_breaks.sgy.reader import SGY
from first_breaks.utils.cuda import get_recommended_device

TRACE_HEADER_NAMES = [v[1] for v in TraceHeaders().headers_schema]


class XAxis(DefaultModel):
    x_axis: Optional[str] = Field(None, description="Source of labels for X axis")

    @field_validator("x_axis")
    def validate_x_axis(cls, v: Any) -> Any:
        if v is None:
            return v
        else:
            if v not in TRACE_HEADER_NAMES:
                raise ValueError(f"'x_axis' must be None or one of trace header name, got {v}")
            else:
                return v


class Encoding(DefaultModel):
    encoding: str = Field("i", description="How to encode value")

    @field_validator("encoding")
    def validate_encoding(cls, v: str) -> str:
        if v not in Headers().format2size.keys():
            raise ValueError(f"'encoding' must be one of {Headers().format2size.keys()}")
        else:
            return v


class TraceHeaderParams(TraceBytePosition, Encoding):
    @field_validator("byte_position")
    def validate_position_depends_on_encoding(cls, v: int, based_validation_info: FieldValidationInfo) -> int:
        encoding = based_validation_info.data["encoding"]
        size = Headers.format2size[encoding]
        if v + size > 240:
            raise ValueError(
                f"'byte_position' is greater than allowed for '{encoding}' encoding. "
                f"Maximum allowed: {240 - size}, got {v}"
            )
        return v


class Source(DefaultModel):
    source: Union[SGY, str, Path, bytes]


class SGYModel(Source):
    sgy: Optional[SGY] = None

    @model_validator(mode="after")  # type: ignore
    def sync_source_and_sgy(self) -> "SGYModel":
        prev_assignment = self.model_config.get("validate_assignment", None)
        self.model_config["validate_assignment"] = False

        if self.sgy is None:
            self.sgy = self.source if isinstance(self.source, SGY) else SGY(self.source)

        self.source = self.sgy

        self.model_config["validate_assignment"] = prev_assignment
        return self  # type: ignore


class Device(DefaultModel):
    device: Literal["cpu", "cuda"] = Field(
        get_recommended_device(), description="Device to compute first breaks"
    )  # type: ignore



import warnings
from abc import abstractmethod
from typing import Any, Dict, Optional, Tuple

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtGui import QCloseEvent, QDoubleValidator
from PyQt5.QtWidgets import (
    QApplication,
    QCheckBox,
    QDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSpinBox,
    QWidget,
)

from first_breaks.const import HIGH_DPI
from first_breaks.data_models.dependent import Device, TraceHeaderParams, XAxis
from first_breaks.data_models.independent import (
    F1F2,
    F3F4,
    Clip,
    FillBlack,
    Gain,
    InvertX,
    InvertY,
    MaximumTime,
    Normalize,
    PicksUnit,
    TNormalize,
    TracesPerGather,
    VSPView,
)
from first_breaks.data_models.initialised_defaults import DEFAULTS
from first_breaks.desktop.bandfilter_widget import QBandFilterWidget
from first_breaks.desktop.byte_encode_unit_widget import QByteEncodeUnitWidget
from first_breaks.desktop.combobox_with_mapping import QComboBoxMapping
from first_breaks.desktop.radioset_widget import QRadioSetWidget
from first_breaks.desktop.utils import QHSeparationLine, set_geometry
from first_breaks.utils.cuda import ONNX_CUDA_AVAILABLE, get_recommended_device

if HIGH_DPI:
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)

warnings.filterwarnings("ignore")


def build_x_axis_mapping() -> Dict[int, Tuple[str, str]]:
    mapping = {0: ("Sequentially", None)}
    headers = [
        "TRACENO",
        "CHAN",
        "SOURCE",
        "CDP",
        "OFFSET",
        "SOU_X",
        "SOU_Y",
        "REC_X",
        "REC_Y",
        "CDP_X",
        "CDP_Y",
        "REC_ELEV",
        "SOU_ELEV",
        "DEPTH",
        "REC_DATUM",
        "SOU_DATUM",
    ]
    mapping.update({idx: (f"Header: {h}", h) for idx, h in enumerate(headers, 1)})  # type: ignore
    return mapping  # type: ignore


X_AXIS_MAPPING = build_x_axis_mapping()


class PlotseisSettings(Gain, Clip, Normalize, XAxis, FillBlack, F1F2, F3F4, VSPView, InvertX, InvertY):
    pass


class PickingSettings(Gain, Clip, Normalize, F1F2, F3F4, MaximumTime, TracesPerGather, Device):
    pass


class PicksFromFileSettings(TraceHeaderParams, PicksUnit):
    pass


def get_value(qline: QLineEdit, minimum: Optional[float] = None, default: Optional[float] = 1.0) -> float:
    value = qline.text()
    qline.setPlaceholderText(str(default))
    if value.lstrip("-").lstrip("+").lstrip(".").lstrip(","):
        value = float(value)
        if minimum:
            value = max(minimum, value)
        return value
    else:
        return default


def default_set_enabled(widget: QWidget, enabled: bool) -> None:
    widget.setEnabled(enabled)


class _Extras:
    @abstractmethod
    def dict(self) -> Dict[str, Any]:
        raise NotImplementedError

    def enable_fields(self) -> None:
        default_set_enabled(self, True)

    def disable_fields(self) -> None:
        default_set_enabled(self, False)


class GainLine(QLineEdit, _Extras):
    changed_signal = pyqtSignal()

    def __init__(self, gain: float = DEFAULTS.gain, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.default = DEFAULTS.gain
        self.setPlaceholderText(str(self.default))
        gain_validator = QDoubleValidator()
        self.setValidator(gain_validator)
        self.setText(str(gain))
        self.textChanged.connect(self.changed_signal)

    def dict(self) -> Dict[str, Any]:
        value = self.text()
        if value.lstrip("-").lstrip("+").lstrip(".").lstrip(","):
            value = float(value)
        else:
            value = self.default
        return {"gain": value}


class ClipLine(QLineEdit, _Extras):
    changed_signal = pyqtSignal()

    def __init__(self, clip: float = DEFAULTS.clip, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.default = DEFAULTS.clip
        self.minimum = 0.1
        self.setPlaceholderText(str(self.default))

        clip_validator = QDoubleValidator()
        clip_validator.setBottom(0.1)
        self.setValidator(clip_validator)
        self.setText(str(clip))
        self.textChanged.connect(self.changed_signal)

    def dict(self) -> Dict[str, Any]:
        value = self.text()
        if value.lstrip("-").lstrip("+").lstrip(".").lstrip(","):
            value = float(value)
            value = max(self.minimum, value)
        else:
            value = self.default

        return {"clip": value}


class NormalizationLine(QComboBoxMapping, _Extras):
    def __init__(self, normalize: TNormalize = DEFAULTS.normalize):
        super().__init__(
            {0: ["Individual traces", "trace"], 1: ["Gather", "gather"], 2: ["Raw", None]},
            current_value=normalize,  # type: ignore
        )

    def dict(self) -> Dict[str, Any]:
        return {"normalize": self.value()}


class BandfilterLine(QWidget, _Extras):
    changed_signal = pyqtSignal()

    def __init__(
        self,
        f1_f2: Optional[Tuple[float, float]] = DEFAULTS.f1_f2,
        f3_f4: Optional[Tuple[float, float]] = DEFAULTS.f3_f4,
    ):
        super().__init__()
        self.bandfilter_toogle = QCheckBox()
        self.bandfilter_toogle.setCheckState(False)
        self.bandfilter_toogle.clicked.connect(self.apply_bandfilter)
        if f1_f2:
            f1, f2 = f1_f2
        else:
            f1, f2 = None, None
        if f3_f4:
            f3, f4 = f3_f4
        else:
            f3, f4 = None, None
        self.bandfilter_widget = QBandFilterWidget(f1=f1, f2=f2, f3=f3, f4=f4, debug=False)
        layout = QHBoxLayout()
        layout.addWidget(self.bandfilter_toogle)
        layout.addWidget(self.bandfilter_widget)
        self.setLayout(layout)
        self.f1_f2 = f1_f2
        self.f3_f4 = f3_f4
        self.bandfilter_widget.enable_freqs_fields()

    def enable_fields(self) -> None:
        self.bandfilter_toogle.setEnabled(True)
        if not self.bandfilter_toogle.isChecked():
            self.bandfilter_widget.enable_freqs_fields()

    def disable_fields(self) -> None:
        self.bandfilter_toogle.setEnabled(False)
        self.bandfilter_widget.disable_freqs_fields()

    def apply_bandfilter(self, checked: bool) -> None:
        if checked:
            freqs = self.bandfilter_widget.validate_and_get_values()
            if freqs is None:
                self.bandfilter_toogle.setChecked(False)
                self.f1_f2 = None
                self.f3_f4 = None
                self.bandfilter_widget.enable_freqs_fields()
            else:
                self.bandfilter_toogle.setChecked(True)
                if freqs[1] is None and freqs[2] is None:
                    self.f1_f2 = None
                else:
                    self.f1_f2 = (freqs[1], freqs[2])
                if freqs[3] is None and freqs[4] is None:
                    self.f3_f4 = None
                else:
                    self.f3_f4 = (freqs[3], freqs[4])
                self.bandfilter_widget.disable_freqs_fields()
                self.changed_signal.emit()
        else:
            self.f1_f2 = None
            self.f3_f4 = None
            self.bandfilter_widget.enable_freqs_fields()
            self.changed_signal.emit()

    def dict(self) -> Dict[str, Any]:
        return {"f1_f2": self.f1_f2, "f3_f4": self.f3_f4}


class WigglesLine(QRadioSetWidget, _Extras):
    def __init__(self, fill_black: Optional[str] = DEFAULTS.fill_black):
        super().__init__(
            {0: ["Left", "left"], 1: ["Right", "right"], 2: ["No fill", None]}, current_value=fill_black, margins=0
        )

    def dict(self) -> Dict[str, Any]:
        return {"fill_black": self.value()}


class XAxisLine(QComboBoxMapping, _Extras):
    def __init__(self, x_axis: Optional[str] = DEFAULTS.x_axis):
        super().__init__(X_AXIS_MAPPING, current_value=x_axis)  # type: ignore

    def dict(self) -> Dict[str, Any]:
        return {"x_axis": self.value()}


class OrientationLine(QWidget, _Extras):
    changed_signal = pyqtSignal()

    def __init__(
        self, vsp_view: bool = DEFAULTS.vsp_view, invert_x: bool = DEFAULTS.invert_x, invert_y: bool = DEFAULTS.invert_y
    ):
        super().__init__()

        self.layout = QHBoxLayout()
        self.setLayout(self.layout)

        self.widgets = [
            ["VSP view", QCheckBox(), vsp_view, "vsp_view"],
            ["Invert X", QCheckBox(), invert_x, "invert_x"],
            ["Invert Y", QCheckBox(), invert_y, "invert_y"],
        ]

        for label, widget, init, _ in self.widgets:
            label = QLabel(label)
            widget.setChecked(init)
            widget.stateChanged.connect(self.changed_signal.emit)
            sub_layout = QHBoxLayout()
            sub_layout.addWidget(widget, alignment=Qt.AlignRight)
            sub_layout.addWidget(label, alignment=Qt.AlignLeft)
            sub_layout.setSpacing(15)
            sub_layout.setContentsMargins(0, 0, 0, 0)
            self.layout.addLayout(sub_layout)

    def dict(self) -> Dict[str, Any]:
        return {k: w.isChecked() for _, w, _, k in self.widgets}


class PicksFromFileLine(QWidget, _Extras):
    toggle_picks_from_file_signal = pyqtSignal(bool)
    export_picks_from_file_settings_signal = pyqtSignal(PicksFromFileSettings)

    def __init__(self, byte_position: int = 1, first_byte: int = 1, encoding: str = "I", picks_unit: str = "mcs"):

        super().__init__()
        self.picks_from_file_toggle = QCheckBox()
        self.picks_from_file_toggle.setCheckState(False)
        # self.picks_from_file_toggle.stateChanged.connect(self.export_picks_from_file_settings)
        self.picks_from_file_toggle.clicked.connect(self.export_picks_from_file_settings)
        self.picks_from_file_widget = QByteEncodeUnitWidget(
            byte_position=byte_position, first_byte=first_byte, encoding=encoding, picks_unit=picks_unit, margins=0
        )
        self.picks_from_file_widget.values_changed_signal.connect(self.update_picks_from_file_settings)
        self.picks_from_file_settings = self.picks_from_file_widget.get_values()

        sub_layout = QHBoxLayout()
        sub_layout.addWidget(self.picks_from_file_toggle)
        sub_layout.addWidget(self.picks_from_file_widget)
        self.setLayout(sub_layout)

    def update_picks_from_file_settings(self, params: Dict[str, Any]) -> None:
        self.picks_from_file_settings = params

    def export_picks_from_file_settings(self, checked: bool) -> None:
        if checked:
            self.picks_from_file_widget.setEnabled(False)
            self.export_picks_from_file_settings_signal.emit(PicksFromFileSettings(**self.picks_from_file_settings))
            self.toggle_picks_from_file_signal.emit(True)
        else:
            self.toggle_picks_from_file_signal.emit(False)
            self.picks_from_file_widget.setEnabled(True)

    def enable_fields(self) -> None:
        self.picks_from_file_toggle.setEnabled(True)
        if not self.picks_from_file_toggle.isChecked():
            self.picks_from_file_widget.setEnabled(True)

    def disable_fields(self) -> None:
        self.picks_from_file_toggle.setEnabled(False)
        self.picks_from_file_widget.setEnabled(False)


class TracesPerGatherLine(QSpinBox, _Extras):
    changed_signal = pyqtSignal()

    def __init__(self, traces_per_gather: int = DEFAULTS.traces_per_gather, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.setMinimum(1)
        self.setValue(traces_per_gather)

    def dict(self) -> Dict[str, Any]:
        return {"traces_per_gather": int(self.text())}


class MaximumTimeLine(QLineEdit, _Extras):
    changed_signal = pyqtSignal()

    def __init__(self, maximum_time: float = DEFAULTS.maximum_time, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        validator = QDoubleValidator()
        validator.setBottom(0.0)
        self.setValidator(validator)
        self.setText(str(maximum_time))

    def dict(self) -> Dict[str, Any]:
        return {"maximum_time": float(self.text())}


class DeviceLine(QComboBoxMapping, _Extras):
    def __init__(self, device: str = DEFAULTS.device):
        if device == "cuda" and ONNX_CUDA_AVAILABLE:
            current_value = device
        else:
            current_value = "cpu"
        super().__init__({0: ["GPU/CUDA", "cuda"], 1: ["CPU", "cpu"]}, current_value=current_value)
        if not ONNX_CUDA_AVAILABLE:
            self.setEnabled(False)

    def dict(self) -> Dict[str, Any]:
        return {"device": self.value()}


class SettingsProcessingWidget(QDialog):
    export_plotseis_settings_signal = pyqtSignal(PlotseisSettings)
    export_picking_settings_signal = pyqtSignal(PickingSettings)
    export_picks_from_file_settings_signal = pyqtSignal(PicksFromFileSettings)
    toggle_picks_from_file_signal = pyqtSignal(bool)
    interrupt_signal = pyqtSignal()

    def __init__(
        self,
        gain: float = 1.0,
        clip: float = 1.0,
        normalize: Optional[str] = "trace",
        f1_f2: Optional[Tuple[float, float]] = None,
        f3_f4: Optional[Tuple[float, float]] = None,
        fill_black: Optional[str] = "left",
        x_axis: Optional[str] = None,
        byte_position: int = 1,
        first_byte: int = 1,
        encoding: str = "I",
        picks_unit: str = "mcs",
        hide_on_close: bool = False,
        vsp_view: bool = False,
        invert_x: bool = False,
        invert_y: bool = True,
        traces_per_gather: int = 12,
        maximum_time: float = 0.0,
        device: str = get_recommended_device(),
    ):
        super().__init__()
        self.hide_on_close = hide_on_close

        self.setWindowTitle("Settings and Processing")
        self.setWindowModality(Qt.ApplicationModal)
        set_geometry(self, width_rel=0.35, height_rel=0.4, fix_size=False, centralize=True)

        self.layout = QGridLayout()
        self.setLayout(self.layout)

        self._separators = [
            [QHSeparationLine("Processing"), 0],
            [QHSeparationLine("View"), 5],
            [QHSeparationLine("External"), 9],
            [QHSeparationLine("NN picking"), 11],
        ]

        for sep, line in self._separators:
            self.layout.addWidget(sep, line, 0, 1, 2)

        self._inputs = [
            ("Gain", GainLine(gain=gain), 1, True),
            ("Clip", ClipLine(clip=clip), 2, True),
            ("Normalization", NormalizationLine(normalize=normalize), 3, True),
            ("Band filter", BandfilterLine(f1_f2=f1_f2, f3_f4=f3_f4), 4, True),
            ("Filling wiggles with color", WigglesLine(fill_black=fill_black), 6, True),
            ("X Axis", XAxisLine(x_axis=x_axis), 7, True),
            ("Orientation", OrientationLine(vsp_view=vsp_view, invert_x=invert_x, invert_y=invert_y), 8, True),
            ("Traces per gather", TracesPerGatherLine(traces_per_gather=traces_per_gather), 12, False),
            ("Maximum time", MaximumTimeLine(maximum_time=maximum_time), 13, False),
            ("Runtime", DeviceLine(device=device), 14, False),
        ]

        for label, widget, line, is_plotseis_param in self._inputs:
            label = QLabel(label)
            if is_plotseis_param:
                widget.changed_signal.connect(self.export_plotseis_settings)  # type: ignore
            self.layout.addWidget(label, line, 0)
            self.layout.addWidget(widget, line, 1)

        picks_from_file_label = QLabel("Show picks from file")
        picks_from_file_widget = PicksFromFileLine(
            byte_position=byte_position, first_byte=first_byte, encoding=encoding, picks_unit=picks_unit
        )
        picks_from_file_widget.toggle_picks_from_file_signal.connect(self.toggle_picks_from_file_signal)
        picks_from_file_widget.export_picks_from_file_settings_signal.connect(
            self.export_picks_from_file_settings_signal
        )
        self.layout.addWidget(picks_from_file_label, 10, 0)
        self.layout.addWidget(picks_from_file_widget, 10, 1)

        self.picking_run = False
        self.run_button = QPushButton("Run picking", self)
        self.run_button.clicked.connect(self.picking_click)
        self.layout.addWidget(self.run_button)
        self.disable_picking()

        self.set_selection_mode()

        self.show()

    def enable_only_visualizations_settings(self) -> None:
        self.setEnabled(True)
        self.disable_picking()

    def enable_picking(self) -> None:
        self.run_button.setEnabled(True)

    def disable_picking(self) -> None:
        self.run_button.setEnabled(False)

    def picking_click(self) -> None:
        self.picking_run = not self.picking_run
        if self.picking_run:
            self.set_picking_mode()
            self.export_picking_settings_signal.emit(PickingSettings(**self.get_settings()))
        else:
            self.interrupt_signal.emit()

    def set_picking_mode(self) -> None:
        for widget in self.findChildren(QWidget):
            if isinstance(widget, _Extras):
                widget.disable_fields()
        self.run_button.setEnabled(True)
        self.run_button.setText("Stop")

    def set_selection_mode(self) -> None:
        self.picking_run = False
        for widget in self.findChildren(QWidget):
            if isinstance(widget, _Extras):
                widget.enable_fields()
        self.run_button.setEnabled(True)
        self.run_button.setText("Run picking")

    def get_settings(self) -> Dict[str, Any]:
        output: Dict[str, Any] = {}
        for _, w, _, _ in self._inputs:
            output.update(w.dict())
        return output

    def export_plotseis_settings(self) -> None:
        settings = self.get_settings()
        print(settings)
        self.export_plotseis_settings_signal.emit(PlotseisSettings(**settings))

    def closeEvent(self, e: QCloseEvent) -> None:
        if self.hide_on_close:
            e.ignore()
            self.hide()
        else:
            e.accept()

    def accept(self) -> None:
        pass


if __name__ == "__main__":
    app = QApplication([])
    window = SettingsProcessingWidget()
    app.exec_()



import click

from first_breaks.desktop.main_gui import run_app

cli_commands = click.Group()

help_desktop_app = "Launch desktop application for picking. 'app' and 'desktop' are aliases"


@cli_commands.command(help=help_desktop_app)
def app() -> None:
    run_app()


@cli_commands.command(help=help_desktop_app)
def desktop() -> None:
    run_app()



from __future__ import annotations

import io
import shutil
import struct
from pathlib import Path
from typing import Any, Dict, Generator, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd

from first_breaks.sgy.headers import FileHeaders, TraceHeaders
from first_breaks.utils.utils import UnitsConverter, calc_hash, chunk_iterable, get_io

SizeHW = Tuple[int, int]


class NotImplementedReader(Exception):
    pass


class InvalidSGY(Exception):
    pass


class SGYInitParamsError(Exception):
    pass


class InvalidSamplesSlice(Exception):
    pass


class SGY:
    fmt2bps = {
        1: 4,  # 4-byte hexadecimal exponent data (IBM single precision floating point)
        2: 4,  # 4-byte, two's complement integer
        3: 2,  # 2-byte, two's complement integer
        4: 4,  # 32-bit fixed point with gain values (Obsolete)
        5: 4,  # 4-byte, IEEE Floating Point
        6: 8,  # 8-byte, IEEE Floating Point
    }

    @property
    def dt(self) -> Union[int, float]:
        return self._dt

    @property
    def ns(self) -> int:
        return self._ns

    @property
    def ntr(self) -> int:
        return self._ntr

    @property
    def dt_mcs(self) -> float:
        return self.dt

    @property
    def dt_ms(self) -> float:
        return self.dt * 1e-3

    @property
    def num_samples(self) -> int:
        return self.ns

    @property
    def num_traces(self) -> int:
        return self.ntr

    @property
    def shape(self) -> SizeHW:
        return self.ns, self.ntr

    @property
    def fs(self) -> float:
        return 1000 / self.dt_ms

    def ms2index(self, ms_value: float) -> int:
        return self.units_converter.ms2index(ms_value)  # type: ignore

    @property
    def max_time_ms(self) -> float:
        return self.num_samples * self.dt_ms

    def __init__(
        self,
        source: Union[str, Path, bytes, np.ndarray],
        dt_mcs: Optional[Union[int, float]] = None,
        general_headers_schema: FileHeaders = FileHeaders(),
        traces_headers_schema: TraceHeaders = TraceHeaders(),
    ):
        self.source = source
        self._dt_mcs_input = dt_mcs

        # data
        self._traces: Optional[np.ndarray] = None
        self._dt: Optional[int] = None
        self._ns: Optional[int] = None
        self._ntr: Optional[int] = None

        # headers for file
        self.general_headers: Optional[Dict[str, Any]] = None
        self._general_headers_schema: FileHeaders = general_headers_schema

        # headers for traces
        self.traces_headers: Optional[pd.DataFrame] = None
        self._traces_headers_raw: Optional[pd.DataFrame] = None
        self._traces_headers_schema: TraceHeaders = traces_headers_schema

        # other
        self._descriptor: Optional[Union[io.BytesIO, io.FileIO]] = None
        self.is_source_ndarray: Optional[bool] = None
        self._endianess: Optional[str] = None
        self._bps: Optional[int] = None
        self._data_fmt: Optional[int] = None
        self._hash_value: Optional[str] = None
        self.units_converter: Optional[UnitsConverter] = None

        self.__init()

    def get_hash(self) -> Optional[str]:
        if self.is_source_ndarray:
            return None
        else:
            if self._hash_value is None:
                self._descriptor = get_io(self.source, mode="rb")
                self._hash_value = calc_hash(self._descriptor)
                self._descriptor.close()
                self._descriptor = None
            return self._hash_value

    def __init(self) -> None:
        if isinstance(self.source, (str, Path, bytes)):
            if self._dt_mcs_input is not None:
                raise SGYInitParamsError("Argument 'dt_mcs' must be empty if SGY created from external sources")
            self._init_from_external()
            self.is_source_ndarray = False
        elif isinstance(self.source, np.ndarray):
            if self._dt_mcs_input is None:
                raise SGYInitParamsError("Argument 'dt_mcs' is required if nd.array is used as input")
            self._init_from_numpy()
            self.is_source_ndarray = True
        else:
            raise SGYInitParamsError("Only `str, Path, bytes, np.ndarray` types are available as input")
        self.units_converter = UnitsConverter(sgy_mcs=self.dt_mcs)

    def _init_from_numpy(self) -> None:
        assert self.source.ndim == 1 or self.source.ndim == 2, "Only arrays are available"  # type: ignore
        if self.source.ndim == 1:  # type: ignore
            self._ntr = 1
        elif self.source.ndim == 2:  # type: ignore
            self._ntr = self.source.shape[1]  # type: ignore
        else:
            raise ValueError("Only 1D or 2D arrays are available")

        self._traces = self.source  # type: ignore
        self._dt = self._dt_mcs_input  # type: ignore
        self._ns = self.source.shape[0]  # type: ignore

    def _init_from_external(self) -> None:
        self._descriptor = get_io(self.source, mode="rb")
        self._read_endianess()
        self._read_general_headers()
        self._read_traces_headers()
        self._scalar_raw_traces_headers()
        self._descriptor.close()
        self._descriptor = None

    def _read_endianess(self) -> None:
        num_bytes = self._descriptor.seek(0, 2)
        if num_bytes < 3600 + 240 + 1:  # at least general headers, 1 trace headers block and byte for 1 sample
            raise InvalidSGY("Invalid structure of SGY file. File is small")

        self._descriptor.seek(3224)
        value = self._descriptor.read(2)

        big = struct.unpack(">H", value)[0]
        little = struct.unpack("<H", value)[0]

        if 1 <= big <= 16 or 1 <= little <= 16:
            self._endianess = ">" if 1 <= big <= 16 else "<"
        else:
            raise InvalidSGY("Invalid endianess of SGY file")

    def _read_general_headers(self) -> None:
        gen_headers = {}
        for pointer, name, fmt in self._general_headers_schema.headers_schema:
            self._descriptor.seek(pointer)
            size = self._general_headers_schema.get_num_bytes(fmt)
            gen_headers[name] = struct.unpack(f"{self._endianess}{fmt}", self._descriptor.read(size))[0]
        self.general_headers = gen_headers

        self._ns = self.general_headers[self._general_headers_schema.ns_name]
        if self._ns < 1:
            raise InvalidSGY("Invalid number of samples")

        self._dt = self.general_headers[self._general_headers_schema.dt_name]
        if self._dt < 0:
            raise InvalidSGY("Invalid time discretization")

        self._data_fmt = self.general_headers[self._general_headers_schema.data_sample_format_name]
        if self._data_fmt < 1 or self._data_fmt > 6:
            raise NotImplementedReader(f"Not supported samples format '{self._data_fmt}'")
        self._bps = self.fmt2bps[self._data_fmt]

        num_bytes = self._descriptor.seek(0, 2)
        self._ntr = int((num_bytes - 3600) / (240 + self._ns * self._bps))
        if num_bytes != (3600 + (240 + self._ns * self._bps) * self._ntr):
            raise InvalidSGY("Invalid number of bytes")

    def _read_traces_headers(self) -> None:
        traces_headers = {}

        for offset, name, fmt in self._traces_headers_schema.headers_schema:
            header = self._read_custom_trace_header_with_existed_descriptor(offset, fmt)
            traces_headers[name] = header

        self._traces_headers_raw = pd.DataFrame(data=traces_headers)

    def _scalar_raw_traces_headers(self) -> None:
        self.traces_headers = self._traces_headers_raw.copy()
        for scalar_from, apply_to_columns in self._traces_headers_schema.scalar_from2apply.items():
            scalar = self._traces_headers_raw[scalar_from].copy()

            scalar[scalar == 0] = 1
            scalar[scalar < 0] = 1 / abs(scalar[scalar < 0])
            self.traces_headers[apply_to_columns] = self.traces_headers[apply_to_columns].apply(lambda x: scalar * x)

    def _read_custom_trace_header_with_existed_descriptor(self, byte_position: int, encoding: str) -> Tuple[Any, ...]:
        size = self._traces_headers_schema.get_num_bytes(encoding)
        buffer = []
        for idx in range(self._ntr):
            pointer = 3600 + (240 + self._ns * self._bps) * idx + byte_position
            self._descriptor.seek(pointer)
            buffer.append(self._descriptor.read(size))
        return struct.unpack(f"{self._endianess}{encoding * self._ntr}", b"".join(buffer))

    def read_custom_trace_header(self, byte_position: int, encoding: str) -> Tuple[Any, ...]:
        self._descriptor = get_io(self.source, mode="rb")
        result = self._read_custom_trace_header_with_existed_descriptor(byte_position, encoding)
        self._descriptor.close()
        self._descriptor = None
        return result

    def replace_traces(self, traces: np.ndarray) -> None:
        if traces.shape == self.shape:
            self._traces = traces

    def read(self, min_sample: Optional[int] = None, max_sample: Optional[int] = None) -> np.ndarray:
        ids = list(range(self.num_traces))
        traces = self.read_traces_by_ids(ids, min_sample, max_sample)
        self.replace_traces(traces)
        return traces

    def get_chunked_reader(
        self, chunk_size: int, min_sample: Optional[int] = None, max_sample: Optional[int] = None
    ) -> Generator[np.ndarray, None, None]:
        chunk_size = min(chunk_size, self.num_traces)
        all_ids = list(range(self.num_traces))

        for ids in chunk_iterable(all_ids, chunk_size):
            yield self.read_traces_by_ids(ids, min_sample, max_sample)

    def read_traces_by_ids(
        self, ids: Sequence[int], min_sample: Optional[int] = None, max_sample: Optional[int] = None
    ) -> np.ndarray:

        if min_sample is not None:
            if min_sample < 0 or not isinstance(min_sample, int):
                raise InvalidSamplesSlice("Invalid minimum slice index")
            min_sample = int(np.clip(min_sample, 0, self.num_samples))
        else:
            min_sample = 0

        if max_sample is not None:
            if max_sample < 1 or not isinstance(max_sample, int):
                raise InvalidSamplesSlice("Invalid maximum slice index")
            max_sample = int(np.clip(max_sample, 0, self.num_samples))
        else:
            max_sample = self.num_samples

        if min_sample >= max_sample:
            raise InvalidSamplesSlice("Minimum slice index is greater or equal to maximum index")

        len_slice = max_sample - min_sample
        ids = [idx for idx in ids if idx < self.num_traces]

        if len(ids) == 0:
            raise ValueError("The requested IDs were not found in the file")

        if self.is_source_ndarray or (self._traces is not None and self._traces.shape == self.shape):
            return self._read_block_ndarray(ids, min_sample, len_slice)
        else:
            return self._read_block_external(ids, min_sample, len_slice)

    def _read_block_ndarray(self, ids: Sequence[int], min_sample: int, length_slice: int) -> np.ndarray:
        return self._traces[min_sample : min_sample + length_slice, ids]

    def _read_block_external(self, ids: Sequence[int], min_sample: int, length_slice: int) -> np.ndarray:
        buffer = []

        self._descriptor = get_io(self.source, mode="rb")
        for idx in ids:
            pointer = 3600 + 240 + (240 + self._ns * self._bps) * idx + min_sample
            self._descriptor.seek(pointer)
            buffer.append(self._descriptor.read(length_slice * self._bps))
        self._descriptor.close()
        self._descriptor = None

        buffer_tr = b"".join(buffer)
        traces = self._read_traces_from_buffer(buffer_tr, (length_slice, len(ids)))
        return traces

    def _read_traces_from_buffer(self, buffer: bytes, shape: SizeHW) -> np.ndarray:
        if self._data_fmt == 1:
            reader_func = self._read_traces_ibm
        elif self._data_fmt == 2:
            reader_func = self._read_traces_4b_compl_int
        elif self._data_fmt == 3:
            reader_func = self._read_traces_2b_compl_int
        elif self._data_fmt == 4:
            raise NotImplementedReader(
                "Not implemented 32-bit fixed point with gain values reader (format 4 of SGY specification)"
            )
        elif self._data_fmt == 5:
            reader_func = self._read_traces_float
        elif self._data_fmt == 6:
            reader_func = self._read_traces_double
        else:
            raise ValueError("Not supported format")

        return reader_func(buffer, shape)

    @staticmethod
    def _read_ibm(ibm: int) -> float:
        """
        Converts an IBM floating point number into IEEE format.
        :param: ibm - 32 bit unsigned integer: unpack('>L', f.read(4))
        """
        if ibm == 0:
            return 0.0
        sign = ibm >> 31 & 0x01
        exponent = ibm >> 24 & 0x7F
        mantissa = (ibm & 0x00FFFFFF) / float(pow(2, 24))
        return (1 - 2 * sign) * mantissa * pow(16, exponent - 64)

    def _read_traces_ibm(self, buffer: bytes, shape: SizeHW) -> np.ndarray:
        reader = np.vectorize(self._read_ibm)
        array = np.ndarray(shape, f"{self._endianess}u4", buffer, order="F")
        return reader(array)

    @staticmethod
    def _read_compl_int(value: int, num_bits: int) -> int:
        return value - int((value << 1) & 2**num_bits)

    def _read_traces_4b_compl_int(self, buffer: bytes, shape: SizeHW) -> np.ndarray:
        reader = np.vectorize(self._read_compl_int)
        array = np.ndarray(shape, f"{self._endianess}u4", buffer, order="F")
        return reader(array, 32)

    def _read_traces_2b_compl_int(self, buffer: bytes, shape: SizeHW) -> np.ndarray:
        reader = np.vectorize(self._read_compl_int)
        array = np.ndarray(shape, f"{self._endianess}u2", buffer, order="F")
        return reader(array, 16)

    def _read_traces_float(self, buffer: bytes, shape: SizeHW) -> np.ndarray:
        return np.ndarray(shape, f"{self._endianess}f4", buffer, order="F")

    def _read_traces_double(self, buffer: bytes, shape: SizeHW) -> np.ndarray:
        return np.ndarray(shape, f"{self._endianess}f8", buffer, order="F")

    def export_sgy_with_picks(
        self,
        output_fname: Union[str, Path],
        picks_in_samples: List[float],
        byte_position: int = 236,
        encoding: Optional[str] = None,
        picks_unit: Optional[str] = "mcs",
    ) -> None:
        assert not self.is_source_ndarray, "Only true SGY can be used for importing picks"
        assert 0 <= byte_position <= 236, "Only 0-236 bytes can be ised for writing"
        assert len(picks_in_samples) == self.num_traces, "Number of traces and picks differs"
        assert picks_unit in ["ms", "mcs", "sample"]

        Path(output_fname).parent.mkdir(exist_ok=True, parents=True)

        if isinstance(self.source, (str, Path)):
            if Path(self.source).resolve() != Path(output_fname).resolve():
                shutil.copyfile(str(self.source), str(output_fname))
        elif isinstance(self.source, bytes):
            with open(output_fname, "wb+") as f_output:
                f_output.write(self.source)
        else:
            raise TypeError("Invalid type of source data")

        if encoding is None:
            encoding = [
                pack_type
                for _, header, pack_type in self._traces_headers_schema.headers_schema
                if header == self._traces_headers_schema.fb_pick_default
            ][0]

        cast_to = float if encoding in ["f", "d"] else int

        if picks_unit == "ms":
            picks = self.units_converter.index2ms(picks_in_samples, cast_to=cast_to)
        elif picks_unit == "mcs":
            picks = self.units_converter.index2mcs(picks_in_samples, cast_to=cast_to)
        else:
            picks = picks_in_samples

        self._descriptor = get_io(output_fname, mode="r+b")
        for idx, pick in enumerate(picks):  # type: ignore
            pointer = 3600 + (240 + self.num_samples * self._bps) * idx + byte_position
            pick_byte = struct.pack(f"{self._endianess}{encoding}", pick)
            self._descriptor.seek(pointer)
            self._descriptor.write(pick_byte)
        self._descriptor.close()



from pathlib import Path
from typing import Any, Dict, Generator, Optional, Tuple, Union

import numpy as np
import onnxruntime as ort

from first_breaks.picking.ipicker import IPicker
from first_breaks.picking.task import Task
from first_breaks.picking.utils import preprocess_gather
from first_breaks.utils.cuda import ONNX_DEVICE2PROVIDER, get_recommended_device
from first_breaks.utils.utils import calc_hash, chunk_iterable, download_model_onnx


class IteratorOfTask:
    gather_key = "gather"
    gather_ids_key = "gather_ids"

    def __init__(self, task: Task):
        self.task = task
        self.idx2gather_ids = {idx: gather_ids for idx, gather_ids in enumerate(self.task.get_gathers_ids())}
        if self.task.normalize == "gather" and len(self.idx2gather_ids) > 1:
            raise AssertionError(
                "'gather' normalization can't be used for picking when number of gathers > 1. "
                "Use 'gather' normalization it only for visualization"
            )

    def __len__(self) -> int:
        return len(self.idx2gather_ids)

    def __getitem__(self, idx: int) -> Dict[str, np.ndarray]:
        gather_ids = self.idx2gather_ids[idx]
        amplitudes = np.array(
            [-1 if idx in self.task.traces_to_inverse else 1 for idx in range(len(gather_ids))], dtype=np.float32
        )
        gather = self.task.sgy.read_traces_by_ids(gather_ids)
        gather = preprocess_gather(
            data=gather,
            gain=self.task.gain,
            clip=self.task.clip,
            normalize=self.task.normalize,
            f1_f2=self.task.f1_f2,
            f3_f4=self.task.f3_f4,
            fs=self.task.sgy.fs,
        )

        gather = amplitudes[None, :] * gather
        gather = gather[: self.task.maximum_time_sample, :]

        return {self.gather_key: gather, self.gather_ids_key: np.array(gather_ids)}

    def get_batch_generator(self, batch_size: int = 1) -> Generator[Dict[str, np.ndarray], None, None]:
        for ids in chunk_iterable(range(len(self)), batch_size):
            gather_batch = []
            gather_ids_batch = []
            for idx in ids:
                item = self[idx]
                gather_batch.append(item[self.gather_key][None, ...])
                gather_ids_batch.append(item[self.gather_ids_key])

            gather_batch = np.stack(gather_batch, axis=0)
            gather_ids_batch = np.hstack(gather_ids_batch)
            batch = {self.gather_key: gather_batch, self.gather_ids_key: gather_ids_batch}
            yield batch


class PickerONNX(IPicker):
    def __init__(
        self,
        model_path: Optional[Union[str, Path]] = None,
        show_progressbar: bool = True,
        device: str = get_recommended_device(),
        batch_size: int = 1,
    ):
        super().__init__(show_progressbar=show_progressbar)
        assert device in ["cpu", "cuda"]

        if model_path is None:
            model_path = download_model_onnx()
        self.model_path = model_path
        self.model_hash = calc_hash(self.model_path)

        self.device = device
        self.batch_size = batch_size

        self.model: Optional[ort.InferenceSession] = None
        self.init_model()

    def init_model(self) -> None:
        sess_opt = ort.SessionOptions()
        sess_opt.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_ALL
        if self.device == "cuda":
            sess_opt = ort.SessionOptions()
            sess_opt.intra_op_num_threads = 2
            sess_opt.inter_op_num_threads = 2
        self.model = ort.InferenceSession(
            str(self.model_path), providers=[ONNX_DEVICE2PROVIDER[self.device]], sess_options=sess_opt
        )

    def change_settings(  # type: ignore
        self, *args: Any, device: Optional[str] = None, batch_size: Optional[int] = None
    ) -> "PickerONNX":
        if args:
            raise ValueError("Use named arguments instead of positional")

        if device and device != self.device:
            self.device = device
            self.init_model()

        if batch_size:
            self.batch_size = batch_size

        return self

    def pick_batch_of_gathers(self, gather: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        assert gather.ndim == 4
        assert all(dim > 0 for dim in gather.shape)
        outputs = self.model.run(None, {"input": gather})
        return outputs[0], outputs[1]

    def process_task(self, task: Task) -> Task:
        task_picks_in_sample = np.zeros(task.sgy.num_traces)
        task_confidence = np.zeros(task.sgy.num_traces)

        task_iterator = IteratorOfTask(task)
        counter_step_finished = 0
        self.callback_processing_started(len(task_iterator))

        for idx, batch in enumerate(task_iterator.get_batch_generator(batch_size=self.batch_size)):
            self.interrupt_if_need()
            data = batch["gather"].astype(np.float32)
            picks, confidence = self.pick_batch_of_gathers(data)

            indices = batch["gather_ids"]
            task_picks_in_sample[indices.flatten()] = picks.flatten()
            task_confidence[indices.flatten()] = confidence.flatten()

            counter_step_finished += len(data)
            self.callback_step_finished(counter_step_finished)

        self.callback_processing_finished()

        task.success = True
        task.picks_in_samples = task_picks_in_sample.astype(int).tolist()
        task.confidence = task_confidence.tolist()
        task.model_hash = self.model_hash

        self.need_interrupt = False

        return task
